package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.Territory;
import games.strategy.engine.gamePlayer.IGamePlayer;
import games.strategy.triplea.ai.AbstractAI;
import games.strategy.triplea.delegate.remote.IAbstractPlaceDelegate;
import games.strategy.triplea.delegate.remote.IMoveDelegate;
import games.strategy.triplea.delegate.remote.IPurchaseDelegate;
import games.strategy.triplea.delegate.remote.ITechDelegate;
import games.strategy.triplea.player.ITripleaPlayer;

/**
 * TestAI
 * 
 * @author A.S.
 * @since 2015
 */

public class TestAI extends AbstractAI implements IGamePlayer, ITripleaPlayer
{	
	private final static SimpleLogger logger = new SimpleLogger();
	
	// Phases
	private TestPurchaseAI testPurchaseAI;
	private TestMoveAI testMoveAI;
	private TestPlaceAI testPlaceAI;
	
	// Analysis of board
	private TestAnalyseBoardAI analyseBoardAI;
	
	// Alignment of the AI player (could either be dynamic, or static and possibly be set in the GUI before the game)
	private int alignmentOfAI = 0;
	
	// Create new TripleAPlayer
	public TestAI(final String name, final String type)
	{
		super(name, type);

		testPurchaseAI = null;
		testMoveAI = null;
		analyseBoardAI = null;
		testPlaceAI = null;
	}
	
	// Note: the implementation of this phase is not very sophisticated yet, and mainly build for testing purposes 
	@Override
	protected void purchase(final boolean purchaseForBid, final int PUsToSpend, final IPurchaseDelegate purchaseDelegate, final GameData data, final PlayerID player)
	{
		analyseBoardAI = new TestAnalyseBoardAI(data, player);
		
		// Compute alignment of AI for this game round for usage in all phases
		alignmentOfAI = analyseBoardAI.getAlignmentOfAIForThisRound();
		
		logger.logMessage("----------------- Start of AI turn -----------------"); // Part of test log: "Main test log"
		logger.logMessage("alignmentOfAI computed: " + alignmentOfAI); // Part of test log: "Main test log"
		
		alignmentOfAI = 2; // hard set for testing purposes
		
		logger.logMessage("alignmentOfAI hard set: " + alignmentOfAI); // Part of test log: "Main test log"
		
		testPurchaseAI = new TestPurchaseAI(purchaseForBid, PUsToSpend, purchaseDelegate, data, player, alignmentOfAI);		
		
		// Execute production
		testPurchaseAI.doPurchase();
	}
	
	// TODO
	@Override
	protected void tech(final ITechDelegate techDelegate, final GameData data, final PlayerID player)
	{
	}
	
	// Note: this phase is the key component of the TestAI gamer
	@Override
	protected void move(final boolean nonCombat, final IMoveDelegate moveDel, final GameData data, final PlayerID player)
	{
		long startTime = System.currentTimeMillis(); // Execution time test
		
		// TODO: check, if it would be sufficient to change only the the game data of the current instance of TestAnalyseBoardAI, instead creating a new instance
		if (nonCombat)
		{
			analyseBoardAI = new TestAnalyseBoardAI(data, player);
		}
		
		testMoveAI = new TestMoveAI(null, null, null, null, null, null, null, nonCombat);
		
		// logger.logMessage("nonCombat: " + nonCombat + "\n" + testMoveAI.getGameStatusAtSelectedNode(data));
		
		testMoveAI.findAndDoMove(moveDel, data, player, alignmentOfAI, analyseBoardAI, this);
		
		long endTime = System.currentTimeMillis(); // Execution time test
		logger.logMessage("wie lange dauert nonCombatMove: " + nonCombat + " insgesamt: " + ((endTime-startTime)/1000) + "sec"); // Part of test log: Main test log
		
		logger.logMessage("------------------------ ENDE nonCombatMove: " + nonCombat + " --------------------------"); // Part of test log: "Main test log"
	}

	// Note: the implementation of this phase is not very sophisticated yet, and mainly build for testing purposes
	@Override
	protected void place(final boolean placeForBid, final IAbstractPlaceDelegate placeDelegate, final GameData data, final PlayerID player)
	{
		testPlaceAI = new TestPlaceAI(placeForBid, placeDelegate, data, player, alignmentOfAI);
		
		testPlaceAI.handlePlace();
	}
	
	// TODO
	@Override
	public boolean shouldBomberBomb(final Territory territory)
	{
		return false;
	}
}