package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.Route;
import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;
import games.strategy.triplea.TripleAUnit;
import games.strategy.triplea.attatchments.TerritoryAttachment;
import games.strategy.triplea.attatchments.UnitAttachment;
import games.strategy.triplea.delegate.Matches;
import games.strategy.util.CompositeMatchAnd;

/**
 * TestAnalyseBoardAI
 * 
 * @author A.S.
 * @since 2015
 */

public class TestAnalyseBoardAI 
{
	private final static SimpleLogger logger = new SimpleLogger();
	
	private GameData data;
	private PlayerID player;
	
	private ArrayList<TestAnalyseBoardContainer> playerStrength;
	
	private double valueForAlignmentDetermination;
	
	private int alignmentOfAIForThisRound;
	
	private ArrayList<TestUnitAndItsTerritoryContainer> unitsThatCanMove;
	private HashMap<String, Territory> territoriesToMoveTo;
	
	private ArrayList<String> goalsList;
	
	public TestAnalyseBoardAI(GameData data, PlayerID player)
	{
		this.data = data;
		this.player = player;
		
		playerStrength = new ArrayList<TestAnalyseBoardContainer>();
		
		valueForAlignmentDetermination = 0;
		alignmentOfAIForThisRound = 0;
		
		unitsThatCanMove = new ArrayList<TestUnitAndItsTerritoryContainer>(); 
		territoriesToMoveTo = new HashMap<String, Territory>(); 
		
		goalsList = new ArrayList<String>();
	}
		
	/*
	 * Compute alignment for this game round.
	 * 
	 * Current implementation: Compare the units and territories of the TestAI player with the units and territories of the enemy players, calculate the relative strength of all players, and
	 * derive the appropriate alignment.
	 */
	public int getAlignmentOfAIForThisRound()
	{	
		double absoluteValueOfBestEnemyPlayer = 0;
		
		for (int x = 0; x < data.getPlayerList().size(); x++)
    	{		
    		int totalUnitCount = 0;
        	int factoryCount = 0;
        	int victoryCitiesCount = 0;
        	int productionUnitsCountPerRound = 0;
        	
			for (Territory t:data.getMap().getTerritoriesOwnedBy(data.getPlayerList().getPlayers().get(x)))
			{	
				totalUnitCount += t.getUnits().getUnitCount((data.getPlayerList().getPlayers().get(x))); 
				
				if (t.getUnits().someMatch(new CompositeMatchAnd<Unit>(Matches.unitIsOwnedBy(data.getPlayerList().getPlayers().get(x)), Matches.UnitCanProduceUnits)))
					factoryCount++;
				
				if (TerritoryAttachment.get(t).getVictoryCity()!=0)
					victoryCitiesCount++;
		
				productionUnitsCountPerRound+=TerritoryAttachment.get(t).getProduction();
				
				// TODO: add reasonable parameters: unit types, units relative to territories,...
			}
			
			TestAnalyseBoardContainer currentPlayer = new TestAnalyseBoardContainer();
			
			currentPlayer.setPlayer(data.getPlayerList().getPlayers().get(x));
			currentPlayer.setTotalUnitCount(totalUnitCount);
			currentPlayer.setFactoryCount(factoryCount);
			currentPlayer.setVictoryCitiesCount(victoryCitiesCount);
			currentPlayer.setTerritoryCount(data.getMap().getTerritoriesOwnedBy(data.getPlayerList().getPlayers().get(x)).size());
			currentPlayer.setProductionUnitsCount(data.getPlayerList().getPlayers().get(x).getResources().getQuantity("PUs"));
			currentPlayer.setProductionUnitsCountPerRound(productionUnitsCountPerRound);
			
			double currentPlayerStrength = 0;
			
			currentPlayerStrength = currentPlayer.getTotalUnitCount()
					+ currentPlayer.getFactoryCount()*1.5
					+ currentPlayer.getProductionUnitsCount()*0.5
					+ currentPlayer.getProductionUnitsCountPerRound()*0.5;
				
			currentPlayer.setPlayerStrength(currentPlayerStrength);
			
			playerStrength.add(currentPlayer);
			
			if (currentPlayerStrength > absoluteValueOfBestEnemyPlayer && !Matches.isAllied(player, data).match(data.getPlayerList().getPlayers().get(x)))
			{
				absoluteValueOfBestEnemyPlayer = currentPlayerStrength;
			}
    	}
		
		for (int x = 0; x < data.getPlayerList().size(); x++)
		{
			double currentPlayerStrengthRelative = 0;
						
			currentPlayerStrengthRelative = playerStrength.get(x).getPlayerStrength()/absoluteValueOfBestEnemyPlayer*100;
			
			playerStrength.get(x).setPlayerStrengthRelative(currentPlayerStrengthRelative);
		}
		
		/*
		for (int x = 0; x < data.getPlayerList().size(); x++)
		{
			logger.logMessage("PlayerName: " + playerStrength.get(x).getPlayer().getName() + " totalUnitCount: " + playerStrength.get(x).getTotalUnitCount() 
					+ " FactoryCount: " + playerStrength.get(x).getFactoryCount() + " VictoryCitiesCount: " + playerStrength.get(x).getVictoryCitiesCount()
					+ " TerritoryCount: " + playerStrength.get(x).getTerritoryCount() + " PlayerStrength: " + playerStrength.get(x).getPlayerStrength()
					+ " PlayerStrengthRelative: " + playerStrength.get(x).getPlayerStrengthRelative());
		}
		*/
		
		for (int x = 0; x < data.getPlayerList().size(); x++)
		{
			if (playerStrength.get(x).getPlayer().equals(player))
			{
				valueForAlignmentDetermination = playerStrength.get(x).getPlayerStrengthRelative();
			}
		}
		
		// logger.logMessage("valueForAlignmentDetermination: " + valueForAlignmentDetermination);
		
		// Determine alignment
		if (valueForAlignmentDetermination <= 60)
		{
			alignmentOfAIForThisRound = 1; // defensive
		}
		else if (valueForAlignmentDetermination > 60 && valueForAlignmentDetermination <= 95)
		{
			alignmentOfAIForThisRound = 2; // normal
		}
		else if (valueForAlignmentDetermination > 95 && valueForAlignmentDetermination <= 100)
		{
			alignmentOfAIForThisRound = 3; // offensive
		}
		else
		{
			logger.logMessage("TABAI: invalid valueForAlignmentDetermination");
		}
		
		return alignmentOfAIForThisRound;
	}
	
	public void setOverallStrategyForThisMove(boolean nonCombat)
	{		
		int alignmentOfAI = getAlignmentOfAIForThisRound();
		
		// hard set of alignment for testing purposes
		alignmentOfAI = 2;
			
		// logger.logMessage("TABAI alignmentOfAI: " + alignmentOfAI);
		
		goalsList.clear();
		
		if (alignmentOfAI == 1)
		{
			if (!nonCombat)
			{
				/*
				 * Add goals for combatMove with defensive alignment
				 */
				goalsList.add("Do nothing");
			}
			else if (nonCombat)
			{
				/*
				 * Add goals for nonCombatMove with defensive alignment
				 */
				goalsList.add("Build defense for own victory cities");
				goalsList.add("Defend marine landing points");		
				goalsList.add("More own units than enemy units in area around own territory");	
			}
			else
			{
				logger.logMessage("invalid value for nonCombat");
			}
		}
		else if (alignmentOfAI == 2)
		{
			if (!nonCombat)
			{
				/*
				 * Add goals for combatMove with normal alignment
				 */
				goalsList.add("Controlled attack of nearest enemy territories");
			}
			else if (nonCombat)
			{
				/*
				 * Add goals for nonCombatMove with normal alignment
				 */
				goalsList.add("Build and supply frontline");
				goalsList.add("Build defense for own victory cities");
				goalsList.add("Defend marine landing points");
			}
			else
			{
				logger.logMessage("invalid value for nonCombat");
			}
		}
		else if (alignmentOfAI == 3)
		{
			if (!nonCombat)
			{
				/*
				 * Add goals for combatMove with offensive alignment
				 */
				goalsList.add("Do nothing");
			}
			else if (nonCombat)
			{
				/*
				 * Add goals for nonCombatMove with offensive alignment
				 */
				goalsList.add("Do nothing");
			}
			else
			{
				logger.logMessage("invalid value for nonCombat");
			}
		}
		else
		{
			logger.logMessage("invalid value for alignmentOfAI");
		}
		
		// Part of test log: "Main test log"
		
		logger.logMessage("TABAI: goalsList: " + goalsList.size());
		
		for (String s:goalsList)
		{
			logger.logMessage("TABAI goals: " + s);
		}
		
		TestSetAndEvaluateGoals setGoals = new TestSetAndEvaluateGoals();
		// logger.logMessage("ttm before set (should be 0): " + territoriesToMoveTo.size());
		territoriesToMoveTo = setGoals.setRelevantTerritoriesForGoals(goalsList, data, player);

		// Part of test log: "territoriesToMoveoTo and unitsThatCanMove"
		logger.logMessage("TABAI: territoriesToMoveTo Size: " + territoriesToMoveTo.size());
		/*
		for(Entry<String, Territory> currentTerritory:territoriesToMoveTo.entrySet())
		{
			logger.logMessage("territoryToMoveTo: " + currentTerritory.getKey());
		}
		*/
		
		unitsThatCanMove.clear();
		
		// Units should be added only once
		boolean addUnitToList = false;
		
		// indexCounter for transports
		int transportsCount = 0;
		
		// Part of test log: "Test, if all units are found"
		/*
		int testSizeUnitsPlayer = 0;
		int testSizeUnitsOverall = 0;
		*/
		
		for (Territory currentTerritory:data.getMap().getTerritories())
		{			
			// Part of test log: "Test, if all units are found"
			/*
			if(currentTerritory.isWater() || currentTerritory.getOwner().equals(player))
			{
				logger.logMessage("Territory: " + currentTerritory.getName() + " units: " + currentTerritory.getUnits().size());
			}
			*/
			
			for (Unit currentUnit:currentTerritory.getUnits())
			{			
				// Part of test log: "Test, if all units are found"
				/*
				testSizeUnitsOverall++;
				
				if(currentUnit.getOwner().equals(player))
				{
					logger.logMessage("currentUnit: " + currentUnit.getID().toString() + " Type: " + currentUnit.getUnitType());
					testSizeUnitsPlayer++;
				}	
				*/
				
				// logger.logMessage("TABAI: currentUnit: " + currentUnit.getID().toString() + " vom Typ " + currentUnit.getUnitType());
				// logger.logMessage("TABAI: currentUnit.getTransportCost(): " + UnitAttachment.get(currentUnit.getUnitType()).getTransportCost());
				// logger.logMessage("TABAI: currentUnit.getTransportCapacity(): " + UnitAttachment.get(currentUnit.getUnitType()).getTransportCapacity());
				
				if (currentUnit.getOwner().equals(player)
						&& UnitAttachment.get(currentUnit.getUnitType()).getMovement(player) > 0
						&& ((nonCombat == UnitAttachment.get(currentUnit.getUnitType()).getCanNotMoveDuringCombatMove())
								|| ((Matches.unitHasMovementLeft.match(currentUnit) && !Matches.UnitWasInCombat.match(currentUnit) && !UnitAttachment.get(currentUnit.getUnitType()).getCanNotMoveDuringCombatMove())) // in non-Combat Einheiten miteinbeziehen, die Movement von combat left und dort nicht gekaempft haben (-> Laut rulebook duerfen Einheiten in nonCombat nicht mehr ziehen, wenn sie zuvor gekaempft haben) 
								|| Matches.UnitIsAir.match(currentUnit)))
				{
					if (nonCombat && Matches.UnitIsAir.match(currentUnit))
					{
						unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));

						break;
					}
					else
					{	
						for(Entry<String, Territory> territoryToMoveTo:territoriesToMoveTo.entrySet())
						{	
							Route r = data.getMap().getRoute(currentTerritory, territoryToMoveTo.getValue());

							addUnitToList = false;

							// General preconditions for unitsThatCanMove
							if ((Matches.UnitHasEnoughMovementForRoute(r).match(currentUnit) 
									|| Matches.UnitIsTransport.match(currentUnit))
									&& Matches.unitHasMovementLeft.match(currentUnit)
									&& !(r.hasLand() && Matches.UnitIsSea.match(currentUnit))
									&& !(r.hasWater() && Matches.UnitIsLand.match(currentUnit))
									&& !(nonCombat && (Matches.territoryHasEnemyLandUnits(player, data).match(territoryToMoveTo.getValue())
											|| Matches.territoryHasEnemySeaUnits(player, data).match(territoryToMoveTo.getValue()))))
							{
								if (Matches.UnitIsTransport.match(currentUnit)
										&& Matches.UnitHasEnoughMovementForRoute(r).match(currentUnit))	
								{
									unitsThatCanMove.add(0, new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
									
									transportsCount++;
									
									addUnitToList = true;
																		
									boolean unitAlreadyInList = false;

									for (Territory step:r.getAllTerritories())
									{
										for (Territory t:data.getMap().getTerritories())
										{	
											for (Unit unit:t.getUnits().getUnits())
											{
												if (unit.getOwner().equals(player) && Matches.UnitIsLand.match(unit)
														&& (nonCombat == UnitAttachment.get(unit.getUnitType()).getCanNotMoveDuringCombatMove()
															|| (Matches.unitHasMovementLeft.match(unit) && !Matches.UnitWasInCombat.match(unit) && !UnitAttachment.get(unit.getUnitType()).getCanNotMoveDuringCombatMove())))
												{
													unitAlreadyInList = false;
													
													Route routeToTransport = new Route();
													
													routeToTransport.setStart(t);
													
													if (!t.toString().equals(step.toString())) // no loops
													{
														routeToTransport.add(step);
													}
		
													if (!t.toString().equals(step.toString()) // no loops
														&& (Matches.UnitHasEnoughMovementForRoute(routeToTransport).match(unit) 
															&& routeToTransport.hasExactlyOneStep()
															&& routeToTransport.allMatch(Matches.TerritoryIsNotImpassable)
															|| (currentTerritory.toString().equals(t.toString())
																	&& Matches.UnitIsLand.match(currentUnit)
																	&& unit.getOwner().equals(player))))
													{	
														for (TestUnitAndItsTerritoryContainer uT:unitsThatCanMove)
														{	
															if (uT.getUnit().getID().toString().equals(unit.getID().toString()))
															{
																unitAlreadyInList = true;
																break;
															}
														}
														
														if (!unitAlreadyInList)
														{
															unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(unit, t));
														}
													}	
												}	
											}
										}
									}	
								}
								else if (Matches.UnitIsTransport.match(currentUnit)
										&& !Matches.UnitHasEnoughMovementForRoute(r).match(currentUnit)
										&& TripleAUnit.get(currentUnit).getMovementLeft() == r.getSteps().size()-1)
								{
									boolean unitAlreadyInList = false;

									for (Territory step:r.getAllTerritories())
									{
										if (!step.toString().equals(r.getEnd().toString())) // = only to the second last territory
										{
											if (!unitAlreadyInList)
											{
												for (Territory neighbour:data.getMap().getNeighbors(step))
												{
													if (territoryToMoveTo.getKey().equals(neighbour.toString()) && !neighbour.isWater())
													{
														for (TestUnitAndItsTerritoryContainer unitInList:unitsThatCanMove)
														{
															if (unitInList.getUnit().getID().toString().equals(currentUnit.getID().toString()))
															{	
																unitAlreadyInList = true;
																break;
															}
														}
														
														if (!unitAlreadyInList)
														{
															unitsThatCanMove.add(0, new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
															
															transportsCount++;
															
															addUnitToList = true;
															
															break;
														}
													}
												}
												
												for (Territory t:data.getMap().getTerritories())
												{	
													for (Unit unit:t.getUnits().getUnits())
													{
														if (unit.getOwner().equals(player) && Matches.UnitIsLand.match(unit)
																&& (nonCombat == UnitAttachment.get(unit.getUnitType()).getCanNotMoveDuringCombatMove()
																	|| (Matches.unitHasMovementLeft.match(unit) && !Matches.UnitWasInCombat.match(unit) && !UnitAttachment.get(unit.getUnitType()).getCanNotMoveDuringCombatMove()))) // = in nonCombat nur Eineiten, die nach combat noch ziehen duerfen
														{
															unitAlreadyInList = false;
															
															Route routeToTransport = new Route();
															
															routeToTransport.setStart(t);
															
															if (!t.toString().equals(step.toString())) // no loops
															{
																routeToTransport.add(step);
															}
		
															if (!t.toString().equals(step.toString()) // no loops
																&& (Matches.UnitHasEnoughMovementForRoute(routeToTransport).match(unit) 
																	&& routeToTransport.hasExactlyOneStep()
																	&& routeToTransport.allMatch(Matches.TerritoryIsNotImpassable)
																	|| (currentTerritory.toString().equals(t.toString())
																			&& Matches.UnitIsLand.match(currentUnit)
																			&& unit.getOwner().equals(player))))
															{	
																for (TestUnitAndItsTerritoryContainer uT:unitsThatCanMove)
																{	
																	if (uT.getUnit().getID().toString().equals(unit.getID().toString()))
																	{
																		unitAlreadyInList = true;
																		break;
																	}
																}
																
																if (!unitAlreadyInList)
																{
																	unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(unit, t));
																}
															}
														}	
													}
												}
											}	
										}	
									}
								}
								else if (Matches.UnitIsCarrier.match(currentUnit)
										&& Matches.UnitHasEnoughMovementForRoute(r).match(currentUnit)
										&& !nonCombat)
								{
									unitsThatCanMove.add(transportsCount, new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
																		
									addUnitToList = true;
								}
								else if (!nonCombat && Matches.UnitIsAir.match(currentUnit))
								{
									unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
									
									addUnitToList = true;
								}
								else if (Matches.UnitIsLand.match(currentUnit))
								{	
									boolean unitAlreadyInList = false;
									
									for (TestUnitAndItsTerritoryContainer unitInList:unitsThatCanMove)
									{
										if (unitInList.getUnit().getID().toString().equals(currentUnit.getID().toString()))
										{	
											unitAlreadyInList = true;
											break;
										}
									}
									
									// land unit could be already in the list because of transport
									if (!unitAlreadyInList)
									{	
										// logger.logMessage("TABAI unitsThatCanMove add: " + currentUnit.getID().toString() + " " + currentTerritory.getName());
										unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
										
										addUnitToList = true;
									}	
								}
								else if (Matches.UnitIsSea.match(currentUnit)
										&& !Matches.UnitCanTransport.match(currentUnit)) 
								{	
									unitsThatCanMove.add(new TestUnitAndItsTerritoryContainer(currentUnit, currentTerritory));
									
									addUnitToList = true;
								}
								if (addUnitToList)
								{	
									break; 
								}
							}
						}	
					}
				}	
			}
		}
		// Part of test log: "Test, if all units are found"
		// logger.logMessage("testSizeUnitsOverall: " + testSizeUnitsOverall + " testSizeUnitsPlayer: " + testSizeUnitsPlayer);
		
		
		// Part of test log: "territoriesToMoveoTo and unitsThatCanMove"
		logger.logMessage("TABAI: unitsThatCanMoveForGoalsSize: " + unitsThatCanMove.size());	
		/*
		logger.logMessage("TABAI: Current sorting of unitsThatCanMove: ");
		
		for(TestUnitAndItsTerritoryContainer u:unitsThatCanMove)
		{
			logger.logMessage("TABAI: unitsThatCanMoveForGoals: " + u.getUnit().getID().toString() + " of type " + u.getUnit().getUnitType());
		}
		*/
	}
	
	/*
	 * Getter for move
	 */
	public ArrayList<String> getOverallStrategyForThisMove()
	{
		return goalsList;
	}
	
	public HashMap<String, Territory> getTerritoriesToMoveTo()
	{		
		return territoriesToMoveTo;
	}
	
	public ArrayList<TestUnitAndItsTerritoryContainer> getUnitsThatCanMove()
	{			
		return unitsThatCanMove;
	}
}