package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Map.Entry;

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.Route;
import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;
import games.strategy.engine.data.UnitType;
import games.strategy.engine.delegate.IDelegateBridge;
import games.strategy.triplea.TripleAUnit;
import games.strategy.triplea.attatchments.TerritoryAttachment;
import games.strategy.triplea.attatchments.UnitAttachment;
import games.strategy.triplea.delegate.BattleDelegate;
import games.strategy.triplea.delegate.DelegateFinder;
import games.strategy.triplea.delegate.Matches;
import games.strategy.triplea.delegate.MoveDelegate;
import games.strategy.triplea.delegate.UndoableMove;
import games.strategy.triplea.delegate.remote.IMoveDelegate;
import games.strategy.util.CompositeMatchAnd;

/**
 * TestMoveAI
 *
 * Implementation of the UCT-algorithm for simulating the movement of units
 * 
 * @author A.S.
 * @since 2015
 */

public class TestMoveAI
{	
	private final static SimpleLogger logger = new SimpleLogger();
	
	static GameData originalData;
	static PlayerID playerCopy;
	static TestAI testAIinstance;
	
	static BattleDelegate battleDelForSimulation;
	static IMoveDelegate moveDelForSimulation;  
	static IDelegateBridge bridgeForSimulation;  
	
	// For UCT-formula
	static Random r;
	static double epsilon;

	// For findAndDoMove()
	private List<TestMoveAI> visited;
	private boolean doSimulation;
	private boolean doNothing;
	
	// For weighting of the already executed moves (evaluateGoal() in TestSetAndEvaluateGoals)
	static int depthBeforeRollOut;
	
	// For simulation strategy (rollOut())
	double valueOfLastNodeOfSimulation;
	
	// Lists for children objects (= leafs) of this class-instance (= this node)
	private ArrayList<TestMoveAI> children;
	private ArrayList<TestMoveAI> childrenForSimulation;
	
	// Stats of this class-instance (= this node)
	private double nVisits;
	private double totValue;
	
	// List of units, that still can be moved for this class-instance (= this node)
	private List<TestUnitAndItsTerritoryContainer> unitsThatCanMove;
	
	/*
	 * Each instance of the class (= each node) has and gets three lists:
	 * 
	 * - units moved
	 * - routes
	 * - transports
	 * 
	 * The lists contain the units and routes, which were selected in the previous nodes. Such that the current node can derive, which own units still can be moved.
	 */	
	private List<String> movedUnitsID; 	
	private List<Route> movedRoutes;
	private List<TestTransportContainer> transports;
	private List<Unit> unitMovedWithTransport;
	private List<TestCarrierContainer> carriers;
	private HashMap<String, Route> landingRoutesOfFighters;
	
	static List<String> movedUnitsIDOfCurrentBestMoveSequence;
	static List<Route> movedRoutesOfCurrentBestMoveSequence;
	static List<TestTransportContainer> transportsOfCurrentBestMoveSequence;
	static List<Unit> unitMovedWithTransportOfCurrentBestMoveSequence;
	static List<TestCarrierContainer> carriersOfCurrentBestMoveSequence;
	static HashMap<String, Route> landingRoutesOfFightersOfCurrentBestMoveSequence;
	
	static double valueOfCurrentBestMoveSequence;
	
	static HashMap<String, Territory> territoriesToMoveTo; 
	
	static int alignmentOfAI;
	
	private boolean nonCombat;
	
	static TestAnalyseBoardAI analyseBoard;
	

	/**
	 * 
	 * @param movedUnitsList
	 * @param analyseBoardAI 
	 * @param nonCombat 
	 * @param route 
	 * @param movedFromList
	 * @param movedToList
	 */
	public TestMoveAI(List<TestUnitAndItsTerritoryContainer> unitsThatCanMoveForThisNode, List<String> movedUnitsIDList, List<Route> movedRoutesList, List<Unit> movedWithTransport, List<TestTransportContainer> transportsList, List<TestCarrierContainer> carriersList, HashMap<String, Route> landingRoutes, boolean nonCombat)
	{
		// long startTime = System.currentTimeMillis(); // Execution time test

		// For select and expand	
        children = new ArrayList<TestMoveAI>();
        childrenForSimulation = new ArrayList<TestMoveAI>();
        
		// Create list for this node
		unitsThatCanMove = new ArrayList<TestUnitAndItsTerritoryContainer>();
        
		movedUnitsID = new ArrayList<String>();
        movedRoutes = new ArrayList<Route>();
        
    	// TODO: check and create only, if transports and/or carriers are an available unit type
		transports = new ArrayList<TestTransportContainer>();
        unitMovedWithTransport = new ArrayList<Unit>();
        
		carriers = new ArrayList<TestCarrierContainer>();
		landingRoutesOfFighters = new HashMap<String, Route>();
		
        // Fill lists for this node
        if (!(unitsThatCanMoveForThisNode == null && movedUnitsIDList == null && movedRoutesList == null && movedWithTransport == null))
        {
        	for(TestUnitAndItsTerritoryContainer currentUnit:unitsThatCanMoveForThisNode)
        		unitsThatCanMove.add(currentUnit);
        	
        	for(String currentUnitID:movedUnitsIDList)
        		movedUnitsID.add(currentUnitID);
        	
        	for(Route current:movedRoutesList)
        		movedRoutes.add(current);
        	
        	for(Unit currentTransport:movedWithTransport)
        	{
        		unitMovedWithTransport.add(currentTransport);
        		// logger.logMessage("unitMovedWithTransport add: " + currentTransport);
        	}
        	
        	if (transportsList != null)
        	{        		
        		/*
        		 * Deep copy
        		 * 
        		 * TODO: quite time consuming -> possible alternatives?
        		 */
        		for(TestTransportContainer currentTransport:transportsList)
        		{
        			TestTransportContainer copyTransport = new TestTransportContainer(currentTransport.getUnitObjectOfTransport());
        			
        			for(Entry<String, Territory> territoryWhereTransportCouldBe:currentTransport.getTerritoriesWhereTransportCouldBe().entrySet())
        			{
        				copyTransport.addTerritoryWhereTransportCouldBe(territoryWhereTransportCouldBe.getValue());
        			}
        			
        			for(Entry<String, Territory> territoryWhereTransportCanUnload:currentTransport.getTerritoriesWhereTransportCanUnload().entrySet())
        			{
        				copyTransport.addTerritoryWhereTransportCanUnload(territoryWhereTransportCanUnload.getValue());
        			}
        			
        			for(Unit loadedUnit:currentTransport.getLoadedUnits())
        			{
        				copyTransport.loadUnit(loadedUnit);
        			}
        			
        			copyTransport.changeLoadedUnitsCosts(currentTransport.getLoadedUnitsCosts());
        			copyTransport.setCanTransportMove(currentTransport.getCanTransportMove());
        			copyTransport.clearRoute();
        			copyTransport.setRoute(currentTransport.getRoute());
        			
        			transports.add(copyTransport);
        		}
        	}
        	
        	if (carriersList != null)
        	{        		
        		/*
        		 * Deep copy
        		 * 
        		 * TODO: quite time consuming -> possible alternatives?
        		 */
        		for (TestCarrierContainer currentCarrier:carriersList)
        		{
        			TestCarrierContainer copyCarrier = new TestCarrierContainer(currentCarrier.getUnitObjectOfCarrier());
        			
    				copyCarrier.setTerritoryWhereCarrierCouldBe(currentCarrier.getTerritoryWhereCarrierCouldBe());
        			
        			for (Unit loadedUnit:currentCarrier.getLoadedUnits())
        			{
        				copyCarrier.loadUnit(loadedUnit);
        			}
        			
        			copyCarrier.changeLoadedUnitsCosts(currentCarrier.getLoadedUnitsCosts());

        			// only own carriers have a route to be copied
        			if (!currentCarrier.getIsCarrierAllied())
        			{
	        			copyCarrier.clearRoute();
	        			
	        			copyCarrier.setRoute(currentCarrier.getRoute());
        			}
        			
        			copyCarrier.setIsCarrierAllied(currentCarrier.getIsCarrierAllied());
        			
        			carriers.add(copyCarrier);
        		}	
        	}
        	
        	if (landingRoutes != null)
        	{
        		/*
        		 * Deep copy
        		 * 
        		 * TODO: quite time consuming -> possible alternatives?
        		 */
        		for (Entry<String, Route> currentLandingRoute:landingRoutes.entrySet())
        		{
        			if (currentLandingRoute.getValue().hasNoSteps())
        			{
        				landingRoutesOfFighters.put(currentLandingRoute.getKey(), originalData.getMap().getRoute(currentLandingRoute.getValue().getStart(), currentLandingRoute.getValue().getStart()));
        			}	
        			else if (currentLandingRoute.getValue().hasSteps())
        			{
        				landingRoutesOfFighters.put(currentLandingRoute.getKey(), originalData.getMap().getRoute(currentLandingRoute.getValue().getStart(), currentLandingRoute.getValue().getEnd()));
        			}
        		}	
        	}
        }
  
		// For simulation strategy
        valueOfLastNodeOfSimulation = 0;
    	
		// For UCT-forumla
		r = new Random();
		epsilon = 1e-6;
		
		// Overall differentiation, if TestMoveAI shall be calculated for units in combat or nonCombatMove
		this.nonCombat = nonCombat;
		
		/*
		logger.logMessage("movedUnitsSize: " + movedUnitsID.size() + " unitsThatCanMoveSize: " + unitsThatCanMove.size());
		logger.logMessage("territoriesMovedFromSize: " + movedFromName.size() + " territoriesMovedToSize: " + movedToName.size()); 
		logger.logMessage("territoriesWithOwnUnitsThatCanMove: " + territoriesWithOwnUnitsThatCanMove.size()); 
		
		long endTime = System.currentTimeMillis(); // Execution time test
		logger.logMessage("Total execution time for constructor: " + (endTime-startTime) + "ms"); // Execution time test
		*/
	}
	
	/*
	 * This method is called in the class TestAI and depicts the starting- and endpoint of all events: in order to run the UCT-algorithm for the rootNode ("the first class instance") and select the best overall move, 
	 * findAndDoMove() calls all other methods of the class TestMoveAI step by step. In the end, the selected move sequence is carried out with the original data and move delegate. 
	 * 
	 *
	 * Afterwards, the whole move (combatMove and nonCombatMove) is completed.
     */
	public void findAndDoMove(IMoveDelegate moveDel, GameData data, PlayerID player, int alignmentOfAI, TestAnalyseBoardAI analyseBoard, TestAI testAI) 
	{	      
		
		originalData = data; // GameDataUtils.cloneGameData(data, true);
		
		// relevant player object and instance of TestAI
		playerCopy = originalData.getPlayerList().getPlayerID(player.getName());
		testAIinstance = testAI;
		
		logger.logMessage("testAIInstance: " + testAIinstance); // Part of test log: "Main test log"
		
		// delegates for simulation
		moveDelForSimulation = new MoveDelegate();
		battleDelForSimulation = DelegateFinder.battleDelegate(originalData); // (BattleDelegate) originalData.getDelegateList().getDelegate("battle");
		
		originalData.getDelegateList().addDelegate(battleDelForSimulation);
		
		/*		
		for (IDelegate d:originalData.getDelegateList())
			logger.logMessage("del: " + d.getName());
		*/
		
		bridgeForSimulation = new DummyDelegateBridge(testAIinstance, playerCopy, originalData);
		moveDelForSimulation.setDelegateBridgeAndPlayer(bridgeForSimulation);
		
		movedUnitsIDOfCurrentBestMoveSequence = new ArrayList<String>();
		movedRoutesOfCurrentBestMoveSequence =  new ArrayList<Route>();
		transportsOfCurrentBestMoveSequence = new ArrayList<TestTransportContainer>();
		unitMovedWithTransportOfCurrentBestMoveSequence = new ArrayList<Unit>();
		carriersOfCurrentBestMoveSequence = new ArrayList<TestCarrierContainer>();
		
		if (!nonCombat)
		{
			landingRoutesOfFightersOfCurrentBestMoveSequence = new HashMap<String, Route>();
		}
		// Part of test log: "Content of static list landingRoutesOfFightersOfCurrentBestMoveSequence in findAndDoMove()"
		/*
		else
		{
			for(Entry<String, Route> landingRoute:landingRoutesOfFightersOfCurrentBestMoveSequence.entrySet())
			{
				logger.logMessage("landingRoute start nonCombat: Unit: " + landingRoute.getKey() + " " + landingRoute.getValue());
			}
		}
		*/
		
		valueOfCurrentBestMoveSequence = 0;
		
		TestMoveAI.alignmentOfAI = alignmentOfAI;
		TestMoveAI.analyseBoard = analyseBoard;
		
		analyseBoard.setOverallStrategyForThisMove(nonCombat);
		
		territoriesToMoveTo = new HashMap<String, Territory>();
		
		for (Entry<String, Territory> currentTerritoryToMoveTo:analyseBoard.getTerritoriesToMoveTo().entrySet())
		{
			territoriesToMoveTo.put(currentTerritoryToMoveTo.getKey(), currentTerritoryToMoveTo.getValue());
		}
		
		// Part of test log: "Test, if all units are found in findAndDoMove()"
		//logger.logMessage("analyseBoard.getUnitsThatCanMove().size(): " + analyseBoard.getUnitsThatCanMove().size());
		
    	for (TestUnitAndItsTerritoryContainer currentUnit:analyseBoard.getUnitsThatCanMove())
    	{	
    		unitsThatCanMove.add(currentUnit);
    		
    		// Part of test log: "Test, if all units are found in findAndDoMove()"
    		//logger.logMessage("currentUnit add first time: " + currentUnit.getUnit().getID().toString() + " Type: " + currentUnit.getUnit().getType() + " Territory: " + currentUnit.getTerritoryOfCurrentUnit().toString());
    	}
    	
    	for (Territory currentTerritory:originalData.getMap().getTerritories())
    	{
    		for (Unit currentUnit:currentTerritory.getUnits().getUnits())
    		{
    			if (Matches.UnitIsAlliedCarrier(playerCopy, originalData).match(currentUnit)
    					&& !Matches.unitOwnedBy(playerCopy).match(currentUnit))
    			{
    				TestCarrierContainer currentCarrier = new TestCarrierContainer(currentUnit);
    			
    				currentCarrier.setIsCarrierAllied(true);
    				currentCarrier.setTerritoryWhereCarrierCouldBe(currentTerritory);
    				
    				int loadCost = 0;

    				for (Unit loadedUnit:currentTerritory.getUnits().getUnits())
    				{	
    					if (TripleAUnit.get(loadedUnit).getTransportedBy() != null)
    					{
	    					if (TripleAUnit.get(loadedUnit).getTransportedBy().getID().toString().equals(currentUnit.toString()))
	    					{
	    						loadCost = UnitAttachment.get(loadedUnit.getUnitType()).getTransportCost();
	    						
	    	    				currentCarrier.changeLoadedUnitsCosts(loadCost);
	    	    				
	    	    				loadCost = 0;
	    					}	
    					}	
    				}
					
					carriers.add(currentCarrier);
    			}
    		}
    	}
    	// Part of test log: "Are allied carriers added correctly in findAndDoMove()"
    	/*
    	for (TestCarrierContainer currentCarrier:carriers)
    	{	
    		logger.logMessage("Current carrier: " + currentCarrier.toString() + " Type: " + currentCarrier.getUnitObjectOfCarrier().getUnitType() 
    				+ " Owner: " + currentCarrier.getUnitObjectOfCarrier().getOwner());
    	}
    	*/
    	
		//logger.logMessage("territoriesWithOwnUnitsThatCanMove().size() findAndDoCombatMove START: " + territoriesWithOwnUnitsThatCanMove.size());
		//logger.logMessage("getUnitsThatCanMove().size() findAndDoCombatMove START: " + getUnitsThatCanMove().size());
		// ---------------------------------------------- //
    	
		int howOften = 0; // for testing purposes
		
		//lastNodesVisited = 0;
		
		doSimulation = true;
		
		doNothing = false;
		
		long startTimeForSimulationBreak = System.currentTimeMillis();
		
		/*
		Note: Starting from here, several further comments and log messages of this class are available in german language only, and may be translated to english in future.
		*/
        while (doSimulation)
        {        	
        	//long startTime = System.currentTimeMillis(); // Execution time test
        	 	
        	selectAction(this);
        	
        	howOften++;
        	
        	if (System.currentTimeMillis() > startTimeForSimulationBreak+15000)
        	{
        		doSimulation = false;
        	}
        	
    		// long endTime = System.currentTimeMillis(); // Execution time test
    		// logger.logMessage("Total execution time for while-doSimulation: " + (endTime-startTime) + "ms" + " simulation runs: " + howOften); // Execution time test
        	
        	// logger.logMessage("--------------------------------------------- End of simulation run ---------------------------------------------");
        }
        
        if (!doNothing)
        {
	    	performMoveWithDelegate(moveDel, movedUnitsIDOfCurrentBestMoveSequence,  movedRoutesOfCurrentBestMoveSequence, transportsOfCurrentBestMoveSequence, unitMovedWithTransportOfCurrentBestMoveSequence);
        		        
	    	originalData = null;
        }	
	}
	
	public void selectAction(TestMoveAI current) 
    {
    	visited = new LinkedList<TestMoveAI>();
    	
    	TestMoveAI cur = current;
    	
		visited.add(cur);
		
		depthBeforeRollOut = 0;
		
		if (depthBeforeRollOut == unitsThatCanMove.size()) {
			doNothing = true;
		}
		
		while (!cur.isLeaf())
        {
            cur = cur.select(false);
                        
			depthBeforeRollOut++;
            
            visited.add(cur);   
        }
		
		if (cur.unitsThatCanMove.size() != 0)
		{	
			//long startTime = System.currentTimeMillis(); // Execution time test
			cur.expand(false);
			//expandCounter++; // Execution time test
			//long endTime = System.currentTimeMillis(); // Execution time test
			//logger.logMessage("Total execution time for one expand: " + (endTime-startTime) + "ms" + " Expand counter: " + expandCounter); // Execution time test

			TestMoveAI newNode = cur.select(false);
			
	        visited.add(newNode);
	        
	        //long startTime2 = System.currentTimeMillis(); // Execution time test
	        double value = rollOut(newNode);
	        //long endTime2 = System.currentTimeMillis(); // Execution time test
			//logger.logMessage("Total execution time for one rollOut: " + (endTime2-startTime2) + "ms"); // Execution time test
			
			for (TestMoveAI node:visited) 
			{			
			    node.updateStats(value);		    
			}		
			
			//logger.logMessage("newNode getStats(): " + newNode.getStats());
			//logger.logMessage("cur.getUnitsThatCanMove().size(): " + cur.getUnitsThatCanMove().size());
		}	
		else
		{	
			doSimulation = false;
			//logger.logMessage("nodeStatsEndeSelectActionDurchlauf");
		}
    }
	
    private TestMoveAI select(boolean selectForSimulation) 
    {
    	//long startTime = System.currentTimeMillis(); // Execution time test
    	TestMoveAI selected = null;
    	
    	double bestValue = Double.NEGATIVE_INFINITY; 

    	if (selectForSimulation)
    	{	
	    	for (TestMoveAI c:childrenForSimulation) 
	        {  
	        	
				double uctValue = c.totValue / (c.nVisits + epsilon) + 1 * Math.sqrt(Math.log(nVisits+1) / (c.nVisits + epsilon))  + r.nextDouble() * epsilon;
				
	            if (uctValue > bestValue) 
	            {
	                selected = c;
	                bestValue = uctValue;
	            }
	            
	          //logger.logMessage("uctValueSimulation: " + uctValue);
	        }
	        //logger.logMessage("child: " + getGameStatusAtSelectedNode(selected.getSimulatedDataOfThisNode()));
    	}
    	else
    	{	
	        for (TestMoveAI c:children) 
	        {
	        	// Simplified test
	        	//double uctValue = c.totValue / (c.nVisits + epsilon); // / c.nVisits;
	        	
	        	double uctValue = c.totValue / (c.nVisits + epsilon) + 0.5 * Math.sqrt(Math.log(nVisits+1) / (c.nVisits + epsilon))  + r.nextDouble() * epsilon;
	        	
	            if (uctValue > bestValue) 
	            {
	                selected = c;
	                bestValue = uctValue;
	            }
	                
	            //logger.logMessage("uctValue: " + uctValue);
	        }
    	
	        //logger.logMessage("child: " + getGameStatusAtSelectedNode(selected.getSimulatedDataOfThisNode()));
    	}  
	       
    	//long endTime = System.currentTimeMillis(); // Execution time test
		//logger.logMessage("Total execution time for select: " + (endTime-startTime) + "ms"); // Execution time test

    	return selected;
    }
    
    /*
     * ---- Simulation to the end node (rollOut) ----
	 */
    public double rollOut(TestMoveAI currentNode)
    {    	
        if (currentNode.unitsThatCanMove.size()!=0)
        {
        	//logger.logMessage("rollOut() start and continue");
        	
        	currentNode.expand(true);
        	
        	TestMoveAI newChildNode = currentNode.select(true);
        	
        	//logger.logMessage("value of selected childNode: " + newChildNode.getStats());
        	
        	rollOut(newChildNode);
        }
        else
        {
        	//logger.logMessage("rollOut() end");
        	
        	performMoveWithDelegate(moveDelForSimulation, currentNode.movedUnitsID, currentNode.movedRoutes, currentNode.transports, currentNode.unitMovedWithTransport);
        	
	    	valueOfLastNodeOfSimulation = getValueForThisNode(moveDelForSimulation.getBridge().getData(), currentNode.getMovedUnits(), null, null);
	
	    	//logger.logMessage("valueOfLastNodeOfSimulation: " + valueOfLastNodeOfSimulation);
	    	//logger.logMessage("--------------------------------------");
	    	
	    	// Part of test log: "Main test log"
	    	/*
	    	logger.logMessage("MOVE valueOfCurrentBestMoveSequence: " + valueOfCurrentBestMoveSequence);
	    	logger.logMessage("MOVE valueOfLastNodeOfSimulation: " + valueOfLastNodeOfSimulation);
	    	*/
	    	
	    	if (valueOfLastNodeOfSimulation > valueOfCurrentBestMoveSequence)
	    	{	    		
	    		valueOfCurrentBestMoveSequence = valueOfLastNodeOfSimulation;
	    		
	    		movedUnitsIDOfCurrentBestMoveSequence.clear();
	    		movedRoutesOfCurrentBestMoveSequence.clear();
	    		transportsOfCurrentBestMoveSequence.clear();
	    		unitMovedWithTransportOfCurrentBestMoveSequence.clear();
	    		carriersOfCurrentBestMoveSequence.clear();
	    		
	    		if (!nonCombat)
	    		{
	    			landingRoutesOfFightersOfCurrentBestMoveSequence.clear();
	    		}
	    		
	    		for (String unitMoved:currentNode.movedUnitsID)
	    		{
	    			movedUnitsIDOfCurrentBestMoveSequence.add(unitMoved);
	    		}
	    		
	    		for (Route routeMoved:currentNode.movedRoutes)
	    		{
	    			movedRoutesOfCurrentBestMoveSequence.add(routeMoved);
	    		}
	    		
	    		for (TestTransportContainer transport:currentNode.transports)
	    		{
	    			transportsOfCurrentBestMoveSequence.add(transport);
	    		}
	    		
	    		for (Unit movedWithTransport:currentNode.unitMovedWithTransport)
	    		{
	    			unitMovedWithTransportOfCurrentBestMoveSequence.add(movedWithTransport);
	    		}
	    		
	    		for (TestCarrierContainer carrier:currentNode.carriers)
	    		{
    				carriersOfCurrentBestMoveSequence.add(carrier);
	    		}
	    		
	    		if (!nonCombat)
	    		{
		    		for (Entry<String, Route> routeToLand:currentNode.landingRoutesOfFighters.entrySet())
		    		{			
		    			if (routeToLand.getValue().hasNoSteps())
	        			{
		    				landingRoutesOfFightersOfCurrentBestMoveSequence.put(routeToLand.getKey(), originalData.getMap().getRoute(routeToLand.getValue().getStart(), routeToLand.getValue().getStart()));
	        			}
	        			else if (routeToLand.getValue().hasSteps())
	        			{
	        				landingRoutesOfFightersOfCurrentBestMoveSequence.put(routeToLand.getKey(), originalData.getMap().getRoute(routeToLand.getValue().getStart(), routeToLand.getValue().getEnd()));
	        			}
		    		}
	    		}		
	    	}
	    	
	    	// logger.logMessage("moveDelForSimulation.getMovesMade().size(): " + moveDelForSimulation.getMovesMade().size());
	    	
	    	while (moveDelForSimulation.getMovesMade().size() != 0)
	    	{
	    		UndoableMove uM = (UndoableMove) moveDelForSimulation.getMovesMade().get(moveDelForSimulation.getMovesMade().size()-1);
	    		
	    		//logger.logMessage("moveDelForSimulation.getMovesMade(): " + moveDelForSimulation.getMovesMade().get(moveDelForSimulation.getMovesMade().size()-1));
	    		
	    		//logger.logMessage("can undo move: " + uM.getcanUndo());
	    		
	    		if (!uM.getcanUndo())
	    		{
	    			//logger.logMessage("can not undo move reason: " + uM.getReasonCantUndo());
	    			
	    			uM.setCantUndo(null);
	    			
	    			//logger.logMessage("can undo move after setCantUndo(null): " + uM.getcanUndo());
	    		}
	    		
	    		moveDelForSimulation.undoMove(moveDelForSimulation.getMovesMade().size()-1);
	    	}	
	    	
	    	//logger.logMessage("moveDelForSimulation.getMovesMade() is Empty: " +  moveDelForSimulation.getMovesMade().isEmpty());
        }
 
        return valueOfLastNodeOfSimulation;
    }
    
    /*
     * ------------------------- Properties of the node -------------------------
     * 
     */
    
    public boolean isLeaf() 
    {
        return children.size() == 0;
    }
    
    public void updateStats(double value) 
    {
    	nVisits++;
        totValue += value;
    }
    
    /*
     * ------------------------- Move simulation and expand -------------------------
     * 
     */
    
    /*
     * Creates the children nodes for the current node
     */ 
    public void expand(boolean expandForSimulation)
    {  	  	
    	//long startTime = System.currentTimeMillis(); // Execution time test
		  		
    	// logger.logMessage("----------------------------------- MOVE Start expand with expandForSimulation: " + expandForSimulation + " -----------------------------------"); // Part of test log: "Main test log"
    	
    	//logger.logMessage("MOVE territoriesWithOwnUnitsThatCanMove size: " + territoriesWithOwnUnitsThatCanMove.size());
    	
    	//logger.logMessage("MOVE Check getUnitsThatCanMoveForGoals(): " + unitsThatCanMove.size()); // Part of test log: "Main test log"
    	
    	//logger.logMessage("MOVE territoriesToMoveTo Size: " + territoriesToMoveTo.size());

    	//logger.logMessage("MOVE territoriesToMoveTo.size(): " + territoriesToMoveTo.size());
    	
		//logger.logMessage("getUnitsThatCanMove().get(0) has been found on this territory and is moved: " + dummy.getID());
    	
    	TestUnitAndItsTerritoryContainer currentUnit = unitsThatCanMove.get(0); 
    	
    	//logger.logMessage("MOVE unitsThatCanMove currentUnit: " + currentUnit.getUnit().getID().toString() + " " + currentUnit.getUnit().getUnitType()); // Part of test log: "Main test log"
    	
    	territoriesToMoveTo.clear();
    	
    	for (Entry<String, Territory> currentTerritoryToMoveTo:analyseBoard.getTerritoriesToMoveTo().entrySet())
		{
			territoriesToMoveTo.put(currentTerritoryToMoveTo.getKey(), currentTerritoryToMoveTo.getValue());
		}
    		
		if (!territoriesToMoveTo.containsKey(currentUnit.getTerritoryOfCurrentUnit().getName().toString()))
		{
			territoriesToMoveTo.put(currentUnit.getTerritoryOfCurrentUnit().getName().toString(), currentUnit.getTerritoryOfCurrentUnit());
			
			//logger.logMessage("MOVE add ttm: " + currentUnit.getTerritoryOfCurrentUnit().getName());
		}
		
		if (!UnitAttachment.get(currentUnit.getUnit().getUnitType()).getIsAir() && !UnitAttachment.get(currentUnit.getUnit().getUnitType()).getIsSea())
		{
			for (TestTransportContainer currentTransport:transports)
			{	
				if ((UnitAttachment.get(currentTransport.getUnitObjectOfTransport().getUnitType()).getTransportCapacity()-currentTransport.getLoadedUnitsCosts()) >= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost())
				{	
					for (Entry<String, Territory> territoryWhereTransportCouldBe:currentTransport.getTerritoriesWhereTransportCouldBe().entrySet())
					{
						Route routeToTransport = originalData.getMap().getRoute(currentUnit.getTerritoryOfCurrentUnit(), territoryWhereTransportCouldBe.getValue());
						
						if (routeToTransport.hasExactlyOneStep()
								&& Matches.UnitHasEnoughMovementForRoute(routeToTransport).match(currentUnit.getUnit()) 
								&& routeToTransport.allMatch(Matches.TerritoryIsNotImpassable))
						{
							if (!territoriesToMoveTo.containsKey(territoryWhereTransportCouldBe.getKey()))
							{
								territoriesToMoveTo.put(territoryWhereTransportCouldBe.getKey(), territoryWhereTransportCouldBe.getValue());
							}
						}
					}
				}	
			}
		}
		
		boolean switchToNextUnit = false;
		
		for (Entry<String, Territory> territoryToMoveTo:territoriesToMoveTo.entrySet())
		{		
			if (switchToNextUnit)
			{
				break;
			}
			
			Route currentRoute = originalData.getMap().getRoute(currentUnit.getTerritoryOfCurrentUnit(), territoryToMoveTo.getValue());
			//logger.logMessage("currentRoute: " + currentRoute.toString());
			
			if (Matches.UnitHasEnoughMovementForRoute(currentRoute).match(currentUnit.getUnit()) 
					&& currentRoute.allMatch(Matches.TerritoryIsNotImpassable)
					&& !(Matches.UnitIsLand.match(currentUnit.getUnit()) && currentRoute.crossesWater())
					&& !(Matches.UnitIsSea.match(currentUnit.getUnit()) && currentRoute.hasLand())
			{
				if (Matches.UnitIsLand.match(currentUnit.getUnit())
						&& !(territoryToMoveTo.getValue().isWater() && !getTerritoriesToGetInTransports().containsKey(territoryToMoveTo.getKey())))
				{		
					boolean enemyTerritory = false;
					
					if (nonCombat && ((!territoryToMoveTo.getValue().isWater() && Matches.isTerritoryEnemy(playerCopy, originalData).match(territoryToMoveTo.getValue())))
							|| (territoryToMoveTo.getValue().isWater()) && Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(territoryToMoveTo.getValue()))
					{
						enemyTerritory = true;
					}
					
					//logger.logMessage("Has enough movement: " + Matches.UnitHasEnoughMovementForRoute(r).match(dummy) + " for " + r.toString());
					
					if (currentRoute.hasSteps())
					{
						for (Territory step:currentRoute.getSteps())
						{
							if (!step.toString().equals(currentRoute.getEnd().toString())
									&& ((Matches.isTerritoryEnemy(playerCopy, originalData).match(step) && !step.isWater())
											|| Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(step)))
							{
								enemyTerritory = true;
							}
						}
					}	
					
					if (!enemyTerritory)
					{
						boolean unitHasMoved = false;
						
						if (getTerritoriesToGetInTransports().containsKey(territoryToMoveTo.getKey()))
						{
							for (TestTransportContainer currentTransport:transports)
					    	{									
								/*
								logger.logMessage("transports.size(): " + transports.size());
								logger.logMessage("currentTransport.getLoadedUnits().size(): " + currentTransport.getLoadedUnits().size());
								logger.logMessage("currentTransport.getUnitObjectOfTransport().getUnitType()).getTransportCapacity(): " + UnitAttachment.get(currentTransport.getUnitObjectOfTransport().getUnitType()).getTransportCapacity());		
								logger.logMessage("currentTransport.getLoadedUnitsCosts(): " + currentTransport.getLoadedUnitsCosts());	
								logger.logMessage("UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost(): " + UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost());
								*/
								
								if (TripleAUnit.get(currentUnit.getUnit()).getTransportedBy() != null)
								{
									if (TripleAUnit.get(currentUnit.getUnit()).getTransportedBy().getID().toString().equals(currentTransport.toString()))
									{								
										unitMovedWithTransport.add(currentTransport.getUnitObjectOfTransport());
											
										Route routeToMoveWithTransportRoute = new Route();
										
										int start = 0;
										
										routeToMoveWithTransportRoute.setStart(currentUnit.getTerritoryOfCurrentUnit());
													
										for (Territory routeOfTransportTerritory:currentTransport.getRoute())
										{
											if (start != 0)
											{
												//logger.logMessage("add Territory: " + routeOfTransportTerritory);
												routeToMoveWithTransportRoute.add(routeOfTransportTerritory);
											}
											
											start++;
										}
										
										performSimulatedMove(currentUnit, routeToMoveWithTransportRoute, expandForSimulation);
										
										unitHasMoved = true;
										
										routeToMoveWithTransportRoute = null;
																				
										currentTransport.copyTerritoriesWhereTransportCanUnload();
										
										//logger.logMessage("currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet().size(): " + currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet().size());
										//logger.logMessage("currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet(): " + currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet());
										
										for (Entry<String, Territory> territoryWhereTransportCanUnload:currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet())
										{											
											enemyTerritory = false;
											
											if (nonCombat && Matches.isTerritoryEnemy(playerCopy, originalData).match(territoryWhereTransportCanUnload.getValue()))
											{
												enemyTerritory = true;
											}
											
											if (!currentUnit.getTerritoryOfCurrentUnit().toString().equals(territoryWhereTransportCanUnload.getKey()) && !enemyTerritory)
											{	
												Route routeToTerritoryWhereTransportCanUnload = new Route();
												
												start = 0;
												
												//logger.logMessage("start territory: " + currentTransport.getRoute().get(0));
												routeToTerritoryWhereTransportCanUnload.setStart(currentTransport.getRoute().get(0));
												
												for(Territory routeOfTransportTerritory:currentTransport.getRoute())
												{
													if(start != 0)
													{
														//logger.logMessage("add territory: " + routeOfTransportTerritory);
														routeToTerritoryWhereTransportCanUnload.add(routeOfTransportTerritory);
													}
													
													start++;
												}
												//logger.logMessage("territoryWhereTransportCanUnload: " + territoryWhereTransportCanUnload.getKey());
												routeToTerritoryWhereTransportCanUnload.add(territoryWhereTransportCanUnload.getValue());
												
												currentTransport.clearTerritoriesWhereTransportCanUnload();
												currentTransport.addTerritoryWhereTransportCanUnload(territoryWhereTransportCanUnload.getValue());
												
												currentTransport.setCanTransportMove(false);
												
												performSimulatedMove(currentUnit, routeToTerritoryWhereTransportCanUnload, expandForSimulation);
												
												unitHasMoved = true;
																								
												currentTransport.clearTerritoriesWhereTransportCanUnload();
												
												for (Entry<String, Territory> currentTerritory:currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet())
												{	
													currentTransport.addTerritoryWhereTransportCanUnload(currentTerritory.getValue());
												}
												
												currentTransport.setCanTransportMove(true);
											}	
										}
										
										unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
										
										switchToNextUnit = true;
									}	
								}
								else if (Matches.TerritoryIsLand.match(currentUnit.getTerritoryOfCurrentUnit())
										&& currentTransport.getTerritoriesWhereTransportCouldBe().containsKey(territoryToMoveTo.getKey()) && currentRoute.hasExactlyOneStep()
										&& (UnitAttachment.get(currentTransport.getUnitObjectOfTransport().getUnitType()).getTransportCapacity()-currentTransport.getLoadedUnitsCosts()) >= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost())		
								{
									Route routeToMoveWithTransportRoute = new Route();
									
									routeToMoveWithTransportRoute.setStart(currentUnit.getTerritoryOfCurrentUnit());
									
									boolean territoryToGetInTransport = false;
									
									for (Territory routeOfTransportTerritory:currentTransport.getRoute())
									{
										if (routeOfTransportTerritory.toString().equals(territoryToMoveTo.getKey()))
										{
											territoryToGetInTransport = true;
										}
										
										if (territoryToGetInTransport)
										{	
											routeToMoveWithTransportRoute.add(routeOfTransportTerritory);
										}
									}
									
									currentTransport.loadUnit(currentUnit.getUnit());
									currentTransport.changeLoadedUnitsCosts(UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost());
									
									unitMovedWithTransport.add(currentTransport.getUnitObjectOfTransport());
																		
									performSimulatedMove(currentUnit, routeToMoveWithTransportRoute, expandForSimulation);
									
									unitHasMoved = true;
									
									routeToMoveWithTransportRoute = null;
									
									currentTransport.copyTerritoriesWhereTransportCanUnload();
									
									for (Entry<String, Territory> territoryWhereTransportCanUnload:currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet())
									{
											
										enemyTerritory = false;
										
										if (nonCombat && Matches.isTerritoryEnemy(playerCopy, originalData).match(territoryWhereTransportCanUnload.getValue()))
										{
											enemyTerritory = true;
										}
										
										if (!currentUnit.getTerritoryOfCurrentUnit().toString().equals(territoryWhereTransportCanUnload.getKey()) && !enemyTerritory)
										{	
											Route routeToTerritoryWhereTransportCanUnload = new Route();
											
											routeToTerritoryWhereTransportCanUnload.setStart(currentUnit.getTerritoryOfCurrentUnit());
											
											territoryToGetInTransport = false;
							
											for(Territory routeOfTransportTerritory:currentTransport.getRoute())
											{
												if (routeOfTransportTerritory.toString().equals(territoryToMoveTo.getKey()))
												{
													territoryToGetInTransport = true;
												}
												
												if (territoryToGetInTransport)
												{	
													routeToTerritoryWhereTransportCanUnload.add(routeOfTransportTerritory);
												}
											}
											
											routeToTerritoryWhereTransportCanUnload.add(territoryWhereTransportCanUnload.getValue());
											
											currentTransport.clearTerritoriesWhereTransportCanUnload();
											currentTransport.addTerritoryWhereTransportCanUnload(territoryWhereTransportCanUnload.getValue());
											
											currentTransport.setCanTransportMove(false);
											
											performSimulatedMove(currentUnit, routeToTerritoryWhereTransportCanUnload, expandForSimulation);
											
											unitHasMoved = true;
											
											currentTransport.clearTerritoriesWhereTransportCanUnload();
											
											for (Entry<String, Territory> currentTerritory:currentTransport.getTerritoriesWhereTransportCanUnloadCopy().entrySet())
											{	
												currentTransport.addTerritoryWhereTransportCanUnload(currentTerritory.getValue());
											}
											
											currentTransport.setCanTransportMove(true);
										}	
									}
									
									currentTransport.changeLoadedUnitsCosts(-UnitAttachment.get(currentUnit.getUnit().getUnitType()).getTransportCost());
									currentTransport.unloadUnit(currentUnit.getUnit());
									
									unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
								}
					    	}	
				    	}
						else if (nonCombat && currentUnit.getTerritoryOfCurrentUnit().isWater() && !unitHasMoved)
						{
							unitMovedWithTransport.add(null);
							
							performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
							
							unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
						}
						else if (!currentUnit.getTerritoryOfCurrentUnit().isWater())
						{
							unitMovedWithTransport.add(null);
							
							performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	

							unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
						}
					}	
				}	
				
				if (UnitAttachment.get(currentUnit.getUnit().getUnitType()).getIsAir())
				{
					if (!Matches.UnitCanLandOnCarrier.match(currentUnit.getUnit()) && !nonCombat)
					{
		    			boolean airCanLand = false;
		    			
    					for (Territory territoryToLand:originalData.getMap().getTerritories())
    					{
    						Route routeToLand = originalData.getMap().getRoute(territoryToMoveTo.getValue(), territoryToLand);
    		    			
    						if (currentRoute.numberOfSteps() + routeToLand.numberOfSteps() <= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getMovement(playerCopy))
    						{
    							if (Matches.TerritoryIsLand.match(territoryToLand)
										&& Matches.airCanLandOnThisAlliedNonConqueredLandTerritory(playerCopy, originalData).match(territoryToLand)
										&& routeToLand.allMatch(Matches.TerritoryIsNotImpassable)
										&& routeToLand.allMatch(Matches.TerritoryAllowsCanMoveAirUnitsOverOwnedLand(playerCopy, originalData)))
								{
									airCanLand = true;
									break;
	    						}	
    						}
    					}
    					
    					if (airCanLand)
    					{	
    						unitMovedWithTransport.add(null);
        					
        					performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
    						
    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
    					}
		    		}
					else if (!Matches.UnitCanLandOnCarrier.match(currentUnit.getUnit()) && nonCombat)
					{							
						for (Territory territoryToLand:originalData.getMap().getTerritories())
						{
							Route routeToLand = originalData.getMap().getRoute(currentUnit.getTerritoryOfCurrentUnit(), territoryToLand);
						
							//logger.logMessage("currentUnit.getTerritoryOfCurrentUnit(): " + currentUnit.getTerritoryOfCurrentUnit() + " territoryToLand: " + territoryToLand); // Part of test log: "Main test log"
							
							if (Matches.TerritoryIsLand.match(territoryToLand)
									&& Matches.airCanLandOnThisAlliedNonConqueredLandTerritory(playerCopy, originalData).match(territoryToLand)
									&& Matches.UnitHasEnoughMovementForRoute(routeToLand).match(currentUnit.getUnit()) 
									&& routeToLand.allMatch(Matches.TerritoryIsNotImpassable)
									&& routeToLand.allMatch(Matches.TerritoryAllowsCanMoveAirUnitsOverOwnedLand(playerCopy, originalData)))
							{
	    						unitMovedWithTransport.add(null);
	    						
	        					performSimulatedMove(currentUnit, routeToLand, expandForSimulation);	
	        					
	    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
							}		
						}
						
						break;
					}
					else if (Matches.UnitCanLandOnCarrier.match(currentUnit.getUnit()))
					{
						if (!nonCombat)
						{
							boolean atLeastOneLandTerritoryToLand = false;
							
    						if (currentRoute.numberOfSteps() == UnitAttachment.get(currentUnit.getUnit().getUnitType()).getMovement(playerCopy))
    						{
    							if (territoryToMoveTo.getValue().isWater())
    							{	
	    							for (TestCarrierContainer currentCarrier:carriers)
									{						
			    						if (UnitAttachment.get(currentCarrier.getUnitObjectOfCarrier().getUnitType()).getCarrierCapacity()-currentCarrier.getLoadedUnitsCosts() >= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost()
			    							&& territoryToMoveTo.getKey().equals(currentCarrier.getTerritoryWhereCarrierCouldBe().getName().toString()))
			    						{
				    						currentCarrier.changeLoadedUnitsCosts(UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost());
			    							
				    						unitMovedWithTransport.add(null);
				    										    						
				    						performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
				    										    						
				    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
				    						
				    						currentCarrier.changeLoadedUnitsCosts(-UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost());
		    							}	
			    					}	
    							}	
    							else
    							{
		    						unitMovedWithTransport.add(null);
		    						
		    						performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
		    						
		    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
    							}	
    						}
    						else if (currentRoute.numberOfSteps() < UnitAttachment.get(currentUnit.getUnit().getUnitType()).getMovement(playerCopy))
	    					{
    							Route routeToLand;
    							
	    						for (Territory territoryToLand:originalData.getMap().getTerritories())
		    					{
	    							routeToLand = originalData.getMap().getRoute(territoryToMoveTo.getValue(), territoryToLand);
	    							
		    						if (currentRoute.numberOfSteps() + routeToLand.numberOfSteps() <= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getMovement(playerCopy))
		    						{
			    						if (Matches.airCanLandOnThisAlliedNonConqueredLandTerritory(playerCopy, originalData).match(territoryToLand)
			    								&& routeToLand.allMatch(Matches.TerritoryIsNotImpassable)
			    								&& routeToLand.allMatch(Matches.TerritoryAllowsCanMoveAirUnitsOverOwnedLand(playerCopy, originalData))
			    								&& !atLeastOneLandTerritoryToLand)
			    						{
			    							atLeastOneLandTerritoryToLand = true;
			    											    						
				    						unitMovedWithTransport.add(null);
				    										    						
				    						performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
				    										    						
				    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
			    						}
		    						}
		    					}
		    					
								for (TestCarrierContainer currentCarrier:carriers)
								{	
									routeToLand = originalData.getMap().getRoute(territoryToMoveTo.getValue(), currentCarrier.getTerritoryWhereCarrierCouldBe());
									
		    						if (currentRoute.numberOfSteps() + routeToLand.numberOfSteps() <= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getMovement(playerCopy)
		    							&& (UnitAttachment.get(currentCarrier.getUnitObjectOfCarrier().getUnitType()).getCarrierCapacity()-currentCarrier.getLoadedUnitsCosts()) >= UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost())
	    							{
			    						if (routeToLand.allMatch(Matches.TerritoryIsNotImpassable)
			    								&& routeToLand.allMatch(Matches.TerritoryAllowsCanMoveAirUnitsOverOwnedLand(playerCopy, originalData)))	
			    						{
			    							currentCarrier.changeLoadedUnitsCosts(UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost());
			    							
		    								landingRoutesOfFighters.put(currentUnit.getUnit().getID().toString(), routeToLand);
			    							
				    						unitMovedWithTransport.add(null);
				    						
				    						performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
				    										    						
				    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
				    						
				    						currentCarrier.changeLoadedUnitsCosts(-UnitAttachment.get(currentUnit.getUnit().getUnitType()).getCarrierCost());
				    						landingRoutesOfFighters.remove(currentUnit.getUnit().getID().toString(), routeToLand);
			    						}	
	    							}	
		    					}	
	    					}	
						}
						else if (nonCombat)
						{
							
							if (landingRoutesOfFightersOfCurrentBestMoveSequence.containsKey(currentUnit.getUnit().getID().toString()))
							{
	    						unitMovedWithTransport.add(null);
	    						
	        					// Unit is moved
	        					//logger.logMessage("landingRoute in nonCombat: " + landingRoutesOfFightersOfCurrentBestMoveSequence.get(currentUnit.getUnit().getID().toString())); // Part of test log: "Main test log"
	        					performSimulatedMove(currentUnit, landingRoutesOfFightersOfCurrentBestMoveSequence.get(currentUnit.getUnit().getID().toString()), expandForSimulation);	
	        					
	    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
							}
							
							for (Territory landingTerritory:originalData.getMap().getTerritories())
							{
								Route routeToLand = originalData.getMap().getRoute(currentUnit.getTerritoryOfCurrentUnit(), landingTerritory);
								
								if (!Matches.isTerritoryEnemy(playerCopy, originalData).match(landingTerritory)
										&& routeToLand.numberOfSteps() <= TripleAUnit.get(currentUnit.getUnit()).getMovementLeft())
								{
		    						unitMovedWithTransport.add(null);
		    								        					
		        					performSimulatedMove(currentUnit, originalData.getMap().getRoute(currentUnit.getTerritoryOfCurrentUnit(), landingTerritory), expandForSimulation);	
		        							    						
		    						unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);
								}
							}
							
							break;
						}
					}
				}
				
				// for seaUnits
				if ((UnitAttachment.get(currentUnit.getUnit().getUnitType()).getIsSea())
					&& territoryToMoveTo.getValue().isWater())
				{			
					boolean enemyUnitsOnRoute = false;
					
					if (nonCombat && Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(territoryToMoveTo.getValue()))
					{
						enemyUnitsOnRoute = true;
					}
					
					//logger.logMessage("cR: " + currentRoute);
					
					if (currentRoute.hasSteps())
					{
						for (Territory step:currentRoute.getSteps())
						{
							if (!step.toString().equals(currentRoute.getEnd().toString())
									&& Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(step))
							{
								//logger.logMessage("enemyUnitsOnRoute");
								enemyUnitsOnRoute = true;
							}
						}
					}		
					
					if (!enemyUnitsOnRoute && !Matches.UnitIsSub.match(currentUnit.getUnit()))
					{
						if (Matches.UnitCanTransport.match(currentUnit.getUnit())
							&& !Matches.UnitTypeIsCarrier.match(currentUnit.getUnit().getUnitType()))
						{	
							TestTransportContainer currentTransport = null;
							boolean isTransportInList = false;
							
							for (TestTransportContainer currentTransportInList:transports)
							{
								if (currentTransportInList.toString().equals(currentUnit.getUnit().getID().toString()))
								{
									currentTransport = currentTransportInList;
									isTransportInList = true;
																		
									boolean unitAlreadyInTransport = false;
									
									if (currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnitCount() > 1)
									{
										for (Unit unitToLoad:currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnits())
										{
											if (currentTransport.getLoadedUnits().size() != 0)
											{
												for (Unit loadedUnit:currentTransport.getLoadedUnits())
												{
													if (loadedUnit.getID().toString().equals(unitToLoad.getID().toString()))
													{	
														unitAlreadyInTransport = true;
														break;
													}
												}	
											}
											
											if (!unitAlreadyInTransport && !UnitAttachment.get(unitToLoad.getUnitType()).getIsAir() && !UnitAttachment.get(unitToLoad.getUnitType()).getIsSea())
											{	
												currentTransport.loadUnit(unitToLoad);
												currentTransport.changeLoadedUnitsCosts(UnitAttachment.get(unitToLoad.getUnitType()).getTransportCost());
											}	
										}
									}
									
									currentTransport.clearTerritoriesWhereTransportCouldBe();
									currentTransport.clearTerritoriesWhereTransportCanUnload();
	
									break;
								}
							}
							
							if (!isTransportInList)
							{	
								currentTransport = new TestTransportContainer(currentUnit.getUnit());
																
								boolean unitAlreadyInTransport = false;
								
								if (currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnitCount() > 1) // > 1, because transport is counted too
								{
									for (Unit unitToLoad:currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnits())
									{
										if (currentTransport.getLoadedUnits().size() != 0)
										{
											for (Unit loadedUnit:currentTransport.getLoadedUnits())
											{
												if (loadedUnit.getID().toString().equals(unitToLoad.getID().toString()))
												{	
													unitAlreadyInTransport = true;
													break;
												}
											}	
										}	
										
										if (!unitAlreadyInTransport && !UnitAttachment.get(unitToLoad.getUnitType()).getIsAir() && !UnitAttachment.get(unitToLoad.getUnitType()).getIsSea())
										{	
											currentTransport.loadUnit(unitToLoad);
											currentTransport.changeLoadedUnitsCosts(UnitAttachment.get(unitToLoad.getUnitType()).getTransportCost());
										}	
									}
								}
								
								transports.add(currentTransport);
							}
							
							for (Territory territoryWhereTransportCouldBe:currentRoute.getAllTerritories())
							{	
								//logger.logMessage("territoryWhereTransportCouldBe: " + territoryWhereTransportCouldBe);
								
								currentTransport.addTerritoryWhereTransportCouldBe(territoryWhereTransportCouldBe);
								
								//logger.logMessage("currentTransport.addTerritoryWhereTransportCouldBe(): " + territoryWhereTransportCouldBe);
								//logger.logMessage("currentTransport.getTerritoriesWhereTransportCouldBe().size(): " + currentTransport.getTerritoriesWhereTransportCouldBe().size());
							}
							
							//logger.logMessage("----------------------------------------------------------");
							
							for (Territory territoryWhereTransportCanUnload:originalData.getMap().getNeighbors(territoryToMoveTo.getValue()))
							{	
								//logger.logMessage("territoryWhereTransportCanUnload: " + territoryWhereTransportCanUnload);
								
								if (!territoryWhereTransportCanUnload.isWater())
								{									
									currentTransport.addTerritoryWhereTransportCanUnload(territoryWhereTransportCanUnload);
									
									//logger.logMessage("currentTransport.addTerritoryWhereTransportCanUnload(): " + territoryWhereTransportCanUnload);
									//logger.logMessage("currentTransport.getTerritoryWhereTransportCanUnload().size(): " + currentTransport.getTerritoriesWhereTransportCanUnload().size());
								}
								else
								{
									//logger.logMessage("territory is not suitable for unloading of land units, and therefore not added");
								}
							}
														
							currentTransport.clearRoute();
							
							currentTransport.setRoute(currentRoute.getAllTerritories());
							
							unitMovedWithTransport.add(currentUnit.getUnit());
							
							performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
							
							unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
						}
						else if (Matches.UnitTypeIsCarrier.match(currentUnit.getUnit().getUnitType()))
						{	
							TestCarrierContainer currentCarrier = null;
							boolean isCarrierInList = false;
							
							for (TestCarrierContainer currentCarrierInList:carriers)
							{
								if (currentCarrierInList.toString().equals(currentUnit.getUnit().getID().toString()))
								{
									currentCarrier = currentCarrierInList;
									isCarrierInList = true;
																		
									boolean unitAlreadyInCarrier = false;
									
									if (currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnitCount() > 1)
									{
										for (Unit unitToLoad:currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnits())
										{
											if (currentCarrier.getLoadedUnits().size() != 0)
											{
												for (Unit loadedUnit:currentCarrier.getLoadedUnits())
												{
													if (loadedUnit.getID().toString().equals(unitToLoad.getID().toString()))
													{	
														unitAlreadyInCarrier = true;
														break;
													}
												}	
											}
											
											if (!unitAlreadyInCarrier && Matches.UnitCanLandOnCarrier.match(unitToLoad)
													&& Matches.isAllied(playerCopy, originalData).match(unitToLoad.getOwner())
													&& !Matches.unitIsOwnedBy(playerCopy).match(unitToLoad))
											{	
												currentCarrier.loadUnit(unitToLoad);
												currentCarrier.changeLoadedUnitsCosts(UnitAttachment.get(unitToLoad.getUnitType()).getCarrierCost());
											}	
										}
									}
									
									currentCarrier.clearTerritoryWhereCarrierCouldBe();
	
									break;
								}
							}
							
							if (!isCarrierInList)
							{	
								currentCarrier = new TestCarrierContainer(currentUnit.getUnit());
																
								boolean unitAlreadyInCarrier = false;
								
								if (currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnitCount() > 1)
								{
									for (Unit unitToLoad:currentUnit.getTerritoryOfCurrentUnit().getUnits().getUnits())
									{
										if (currentCarrier.getLoadedUnits().size() != 0)
										{
											for (Unit loadedUnit:currentCarrier.getLoadedUnits())
											{
												if (loadedUnit.getID().toString().equals(unitToLoad.getID().toString()))
												{	
													unitAlreadyInCarrier = true;
													break;
												}
											}	
										}
										
										if (!unitAlreadyInCarrier && Matches.UnitCanLandOnCarrier.match(unitToLoad)
												&& Matches.isAllied(playerCopy, originalData).match(unitToLoad.getOwner())
												&& !Matches.unitIsOwnedBy(playerCopy).match(unitToLoad))
										{	
											currentCarrier.loadUnit(unitToLoad);
											currentCarrier.changeLoadedUnitsCosts(UnitAttachment.get(unitToLoad.getUnitType()).getCarrierCost());
											// logger.logMessage("Create carrier loadUnit: " + unitToLoad.getID().toString() + " Type: " + unitToLoad.getUnitType().toString() + " loadedUnitCosts " + currentCarrier.getLoadedUnitsCosts());
										}	
									}
								}
								
								carriers.add(currentCarrier);
							}
							
							//logger.logMessage("territoryWhereCarrierCouldBe: " + currentRoute.getEnd().getName().toString());
							
							//logger.logMessage("carriers.size(): " + carriers.size());
							//logger.logMessage("currentCarrier: " + currentCarrier.getUnitObjectOfCarrier().getID().toString() + " Typ: " + currentCarrier.getUnitObjectOfCarrier().getUnitType());
							//logger.logMessage("currentRoute.getAllTerritories().size(): " + currentRoute.getAllTerritories().size());
							
							if (currentRoute.getAllTerritories().size() == 1)
							{
								currentCarrier.setTerritoryWhereCarrierCouldBe(currentUnit.getTerritoryOfCurrentUnit());
							}
							else if (currentRoute.getAllTerritories().size() > 1 
									&& !Matches.territoryHasEnemyUnits(playerCopy, originalData).match(currentRoute.getAllTerritories().get(currentRoute.getAllTerritories().size()-1)))
							{
								currentCarrier.setTerritoryWhereCarrierCouldBe(currentRoute.getEnd());
								
								//logger.logMessage("currentCarrier.setTerritoryWhereCarrierCouldBe(): " + territoryWhereCarrierCouldBe);
							}	
							else
							{
								//logger.logMessage("Territory is not suitable for a safe landing of fighters, because enemy units are located on it");
							}
														
							currentCarrier.clearRoute();
							
							/*
							for(Territory t:currentRoute.getAllTerritories())
							{
								logger.logMessage("terr: " + t.toString());
							}
							*/
							
							currentCarrier.setRoute(currentRoute.getAllTerritories());
							
							//logger.logMessage("currentCarrier.getTerritoryWhereCarrierCouldBe(): " + currentCarrier.getTerritoryWhereCarrierCouldBe());
														
							unitMovedWithTransport.add(null);
							
							performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
							
							unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
						}
						// other seaUnits except submarines
						else if (Matches.UnitIsSea.match(currentUnit.getUnit()) 
								&& !Matches.UnitIsCarrier.match(currentUnit.getUnit())
								&& !Matches.UnitIsTransport.match(currentUnit.getUnit())
								&& !Matches.UnitIsSub.match(currentUnit.getUnit()))
						{
							enemyUnitsOnRoute = false;
							
							if (nonCombat && Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(territoryToMoveTo.getValue()))
							{
								enemyUnitsOnRoute = true;
							}
							
							//logger.logMessage("cR: " + currentRoute);
							
							if (currentRoute.hasSteps())
							{
								for (Territory step:currentRoute.getSteps())
								{
									if (!step.toString().equals(currentRoute.getEnd().toString())
											&& Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(step))
									{
										//logger.logMessage("enemyUnitsOnRoute");
										enemyUnitsOnRoute = true;
									}
								}
							}		
							
							if (!enemyUnitsOnRoute)
							{							
								unitMovedWithTransport.add(null);
																
								performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
																
								unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);		
							}
						}
		    		}
					
					if (Matches.UnitIsSub.match(currentUnit.getUnit()))
					{
						enemyUnitsOnRoute = false;
					
						if (currentRoute.hasSteps())
						{
							for (Territory step:currentRoute.getSteps())
							{
								if (!step.toString().equals(currentRoute.getEnd().toString())
										&& step.getUnits().someMatch(Matches.UnitIsDestroyer))
								{
									//logger.logMessage("enemyDestroyerOnRoute");
									enemyUnitsOnRoute = true;
								}
							}
						}	
						
						if ((!nonCombat && !enemyUnitsOnRoute) 
								|| (nonCombat && !enemyUnitsOnRoute && !Matches.territoryHasEnemySeaUnits(playerCopy, originalData).match(territoryToMoveTo.getValue())))
						{						
							unitMovedWithTransport.add(null);
														
							performSimulatedMove(currentUnit, currentRoute, expandForSimulation);	
														
							unitMovedWithTransport.remove(unitMovedWithTransport.size()-1);	
						}	
					}
				}	
			}	
			else
			{
				//logger.logMessage("Unit can not move this route");
			}
		}
		
		//logger.logMessage("ExpandForSimulation " + expandForSimulation + " finished"); // Part of test log: "Main test log"
				
    	//long startTime2 = System.currentTimeMillis(); // Execution time test
			
		//long endTime = System.currentTimeMillis(); // Execution time test
		//logger.logMessage("Total execution time for move simualtion and expand: " + (endTime-startTime) + "ms"); // Execution time test
    }
    
    private void performSimulatedMove(TestUnitAndItsTerritoryContainer unitToMove, Route routeToMove, boolean expandForSimulation)
    {
    	unitsThatCanMove.remove(unitToMove);	
    	
    	movedUnitsID.add(unitToMove.getUnit().getID().toString());
    	//logger.logMessage("pSM: " + routeToMove);
		movedRoutes.add(routeToMove);
		
		if (expandForSimulation) 
		{
			//logger.logMessage("expandForSimulation");
			
    		childrenForSimulation.add(new TestMoveAI(unitsThatCanMove, movedUnitsID, movedRoutes, unitMovedWithTransport, transports, carriers, landingRoutesOfFighters, nonCombat));
		}
		else if (!expandForSimulation)
		{
			//logger.logMessage("expand");
    		children.add(new TestMoveAI(unitsThatCanMove, movedUnitsID, movedRoutes, unitMovedWithTransport, transports, carriers, landingRoutesOfFighters, nonCombat));
		}
		else
		{
			logger.logMessage("Expand could not be done");
		}
		
		unitsThatCanMove.add(0, unitToMove);
		
		movedUnitsID.remove(movedUnitsID.size()-1);
		movedRoutes.remove(movedRoutes.size()-1);
    }
    
    private void performMoveWithDelegate(IMoveDelegate moveDel, List<String> pmMovedUnitsID, List<Route> pmRoutes, List<TestTransportContainer> pmTransports, List<Unit> pmUnitMovedWithTransport)
    {    	
    	List<Unit> unitsMoved = new ArrayList<Unit>();
    	
    	for (String sU:pmMovedUnitsID)
    	{
    		for (Unit u:originalData.getUnits().getUnits())
    		{
    			if (sU.equals(u.getID().toString()))
    			{
    				unitsMoved.add(u);
    			}
    		}
    	}
    	
    	//logger.logMessage("unitsMoved: " + unitsMoved.size());
    	
    	int indexHelper = 0;
    	
    	// logger.logMessage("----------- Execution of the moves related to transports -----------"); // Part of test log: "performSimulatedMove"
    	
    	for (TestTransportContainer currentTransport:pmTransports)
    	{     
    		List<Unit> currentlyMovedUnit = new ArrayList<Unit>();
			Route transporterRoute = new Route();
    		List<Unit> currentlyMovedUnitMovedWithTransport = new ArrayList<Unit>();
    		
    		int listIndex = 0;
    		indexHelper = 0;
			Unit transport = null;
			
			LinkedHashMap<Unit, Integer> transportedUnits = new LinkedHashMap<Unit, Integer>();
			
			//logger.logMessage("unitMovedWithTransportOfCurrentBestMoveSequence: " + unitMovedWithTransportOfCurrentBestMoveSequence.size());
			
    		for (String sU:pmMovedUnitsID)
        	{
    			if (sU.equals(currentTransport.getUnitObjectOfTransport().getID().toString()))
    			{
    				listIndex = indexHelper;
    				
    				transport = currentTransport.getUnitObjectOfTransport();
    				
    				break;
    			}
    			
    			indexHelper++;
        	}
    		
    		transporterRoute = pmRoutes.get(listIndex);
    		    		
    		indexHelper = 0;
    		
    		for (String sU:pmMovedUnitsID)
        	{        			
    			for (Unit loadedUnit:currentTransport.getLoadedUnits())
            	{
        			if (sU.equals(loadedUnit.getID().toString()))
        			{	        				
        				transportedUnits.put(loadedUnit, indexHelper);
        			}	
            	}	
    			
    			indexHelper++;
        	}
    		        		
    		// check and perform moves for units, that want to enter the transport on its starting territory
    		for (Entry<Unit, Integer> transportedUnit:transportedUnits.entrySet())
    		{
    			//logger.logMessage("movedRoutesOfCurrentBestMoveSequence.get(transportedUnit.getValue()).getAllTerritories().get(1).toString(): " + movedRoutesOfCurrentBestMoveSequence.get(transportedUnit.getValue()).getAllTerritories().get(1).toString()); 
    			//logger.logMessage("transporterRoute.getStart().toString(): " + transporterRoute.getStart().toString());
    			
    			if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().size() >= 2)
    			{
    				if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(1).toString().equals(transporterRoute.getStart().toString())
    						&& !pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(0).toString().equals(transporterRoute.getStart().toString()))
    				{
    					Route currentRoute = new Route();
    					
    					currentRoute.setStart(pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(0));
    					currentRoute.add(transporterRoute.getStart());
    					
    					currentlyMovedUnit.add(transportedUnit.getKey());
    					currentlyMovedUnitMovedWithTransport.add(transport);
    					
    					try
    	    			{
    	    				originalData.acquireWriteLock();
    	    				
        					// logger.logMessage("Perform move: Load unit " + transportedUnit.getKey().getID().toString() + " on starting territory " + transporterRoute.getStart().toString() + " of route of the transport: " + currentRoute); // Teil von Testlog: performSimulatedMove
    	    				
        					moveDel.move(currentlyMovedUnit, currentRoute, currentlyMovedUnitMovedWithTransport);
	    					
	    					// logger.logMessage("moveDel: " + moveDel.getMovesMade()); // Part of test log: "performSimulatedMove"
    	    			}
    		    		catch (final Throwable thr)
    	    			{
    		    			logger.logMessage("Error trying to move with moveDel" + thr);
    	    				return;
    	    			}
    	    			finally
    	    			{
    	    				originalData.releaseWriteLock();
    	    			}
    					    					
    					currentRoute = null;
    					
    					currentlyMovedUnit.clear();
    					currentlyMovedUnitMovedWithTransport.clear();
    				}	
				}	
    		}
    		    		
    		if (transporterRoute.hasSteps())
    		{
    			//logger.logMessage("going through the further route of the transport step by step, and load or unload units (if they have reached their target territory on their desired route)"); // Part of test log: "performSimulatedMove"
    			
    			int transportRouteSteps = 0;
				    			
    			ArrayList<Boolean> unitsLoadedInDuringMovement = new ArrayList<Boolean>();
    			int index = 0;
    			
	    		for (Territory step:transporterRoute.getSteps())
	    		{	    			
	    			Route currentRouteTransport = new Route();
	    			
	    			currentRouteTransport.setStart(transporterRoute.getAllTerritories().get(transportRouteSteps));
	    			currentRouteTransport.add(step);
	    			
	    			for (Entry<Unit, Integer> transportedUnit:transportedUnits.entrySet())
					{
	        			if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().size() >= 2)
	        			{
	        				if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(1).toString().equals(transporterRoute.getStart().toString()))
							{
								currentlyMovedUnit.add(transportedUnit.getKey());
								
								//logger.logMessage("currentlyMovedUnit.add(transportedUnit.getKey(): " + transportedUnit.getKey().getID());
							}
	        			}
	        			else if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(0).toString().equals(transporterRoute.getStart().toString()))
						{
							currentlyMovedUnit.add(transportedUnit.getKey());
							
							//logger.logMessage("currentlyMovedUnit.add(transportedUnit.getKey(): " + transportedUnit.getKey().getID());
						}
						
						if (unitsLoadedInDuringMovement.size() != 0)
						{
							if (unitsLoadedInDuringMovement.get(index))
							{
								currentlyMovedUnit.add(transportedUnit.getKey());
								
								//logger.logMessage("currentlyMovedUnit.add(transportedUnit.getKey(): " + transportedUnit.getKey().getID());
								
								//logger.logMessage("unitsLoadedInDuringMovement.get(index): " + unitsLoadedInDuringMovement.get(index));
							}
						}	
						
						index++;
					}
	    			
	    			unitsLoadedInDuringMovement.clear();
	    			index = 0;
	    			
					currentlyMovedUnit.add(transport);
			
					try
	    			{
	    				originalData.acquireWriteLock();
	    				
	    				//logger.logMessage("Transport " + transport.getID().toString() + " of type: " + transport.getUnitType()  
	    						//+ " moves with possibly already loaded units on route " + currentRouteTransport + " one territory forward."); // Part of test log: "performSimulatedMove"
	    				
	    				moveDel.move(currentlyMovedUnit, currentRouteTransport);
	    				
	    				// logger.logMessage("moveDel: " + moveDel.getMovesMade()); // Part of test log: "performSimulatedMove"
	    			}
		    		catch(final Throwable thr)
	    			{
		    			logger.logMessage("Error trying to move with moveDel" + thr);
	    				return;
	    			}
	    			finally
	    			{
	    				originalData.releaseWriteLock();
	    			}
										
					currentRouteTransport = null;
					
					currentlyMovedUnit.clear();
					
	    			for (Entry<Unit, Integer> transportedUnit:transportedUnits.entrySet())
	        		{
	        			if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().size() >= 2)
	        			{
		    				if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(1).toString().equals(step.toString())
		    						&& !pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(0).isWater())
		    				{
		    					Route currentRoute = new Route();
		    					
		    					currentRoute.setStart(pmRoutes.get(transportedUnit.getValue()).getAllTerritories().get(0));
		    					
		    					// logger.logMessage("Land unit " + transportedUnit.getKey().getID().toString() + " of type: " + transportedUnit.getKey().getUnitType() +  " wants to enter transport"); // Part of test log: "performSimulatedMove"
			    					
		    					currentRoute.add(step);
		    					
		    					currentlyMovedUnit.add(transportedUnit.getKey());
		    					currentlyMovedUnitMovedWithTransport.add(transport);
		    					
		        				unitsLoadedInDuringMovement.add(true);
		    						    					
		    					try
		    	    			{
		    	    				originalData.acquireWriteLock();
		    	    
		    	    				moveDel.move(currentlyMovedUnit, currentRoute, currentlyMovedUnitMovedWithTransport);
		    	    				
		    	    				// logger.logMessage("moveDel: " + moveDel.getMovesMade()); // Part of test log: "performSimulatedMove"
		    	    			}
		    		    		catch (final Throwable thr)
		    	    			{
		    		    			logger.logMessage("Error trying to move with moveDel" + thr);
		    	    				return;
		    	    			}
		    	    			finally
		    	    			{
		    	    				originalData.releaseWriteLock();
		    	    			}
			    						    					
		    					currentRoute = null;
		    					
		    					currentlyMovedUnit.clear();
		    					currentlyMovedUnitMovedWithTransport.clear();	    				
		    				}
		    				else
		    				{
		    					unitsLoadedInDuringMovement.add(false);
		    				}
	        			}	
	    				else
	    				{
	    					unitsLoadedInDuringMovement.add(false);
	    				}
	        		}
	    			
	    			for (Entry<Unit, Integer> transportedUnit:transportedUnits.entrySet())
	        		{
	        			if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().size() >= 2)
	        			{
	    					for (Territory neighbour:originalData.getMap().getNeighbors(step))
	    					{
	    						if (neighbour.toString().equals((pmRoutes.get(transportedUnit.getValue()).getEnd().toString()))
	    								&& step.toString().equals(transporterRoute.getEnd().toString())) // -> wenn der Transporter auch schon fertig mit seinem Zug ist
								{
									//logger.logMessage("Land unit " + transportedUnit.getKey().getID().toString() + " of type: " + transportedUnit.getKey().getUnitType() 
										//+ " wants to leave transport and therefore gets unloaded on territory: " + neighbour.toString()); // Part of test log: "performSimulatedMove"
	    							
			    					Route currentRoute = new Route();
			    					
			    					currentRoute.setStart(step);
			    					currentRoute.add(neighbour);
			    					
			    					currentlyMovedUnit.add(transportedUnit.getKey());
			    					
			    					try
			    	    			{
			    	    				originalData.acquireWriteLock();
			    	    
			    	    				moveDel.move(currentlyMovedUnit, currentRoute);
			    	    				
			    	    				//logger.logMessage("moveDel: " + moveDel.getMovesMade()); // Part of test log: "performSimulatedMove"
			    	    			}
			    		    		catch (final Throwable thr)
			    	    			{
			    		    			logger.logMessage("Error trying to move with moveDel" + thr);
			    	    				return;
			    	    			}
			    	    			finally
			    	    			{
			    	    				originalData.releaseWriteLock();
			    	    			}
			    								    					
			    					currentRoute = null;
			    					
			    					currentlyMovedUnit.clear();
			    					currentlyMovedUnitMovedWithTransport.clear();
	    							
	    							break;
								}
	    					}
	        			}	
	        		}	
	    			
	    			transportRouteSteps++;
	    		}
    		}
			else
			{
				for (Entry<Unit, Integer> transportedUnit:transportedUnits.entrySet())
        		{
					for (Territory neighbour:originalData.getMap().getNeighbors(currentTransport.getRoute().get(0)))
					{
						if (pmRoutes.get(transportedUnit.getValue()).getAllTerritories().size() >= 2)
						{
							if (neighbour.toString().equals((pmRoutes.get(transportedUnit.getValue()).getEnd().toString())))
							{
								//logger.logMessage("Land unit " + transportedUnit.getKey().getID().toString() + " of type: " + transportedUnit.getKey().getUnitType() 
										//+ " wants to leave transport and therefore gets unloaded on territory: " + neighbour.toString()); // Part of test log: "performSimulatedMove"
								
		    					Route currentRoute = new Route();
		    					
		    					currentRoute.setStart(currentTransport.getRoute().get(0));
		    					currentRoute.add(neighbour);
		    					
		    					currentlyMovedUnit.add(transportedUnit.getKey());
		    					
		    	        		if (currentRoute.hasSteps())
		    	        		{
			    					try
			    	    			{
			    	    				originalData.acquireWriteLock();
			    	    				
			    	    				moveDel.move(currentlyMovedUnit, currentRoute);
			    	    				
			    	    				//logger.logMessage("moveDel: " + moveDel.getMovesMade()); // Part of test log: "performSimulatedMove"
			    	    			}
			    		    		catch (final Throwable thr)
			    	    			{
			    		    			logger.logMessage("Error trying to move with moveDel" + thr);
			    	    				return;
			    	    			}
			    	    			finally
			    	    			{
			    	    				originalData.releaseWriteLock();
			    	    			}
		    	        		}
			    						    					
		    					currentRoute = null;
		    					
		    					currentlyMovedUnit.clear();
		    					currentlyMovedUnitMovedWithTransport.clear();
								
								break;
							}
						}	
	        		}	
				}
			}
    		
        	// logger.logMessage("-------------- All moves related to this transport have been performed --------------"); // Part of test log: "performSimulatedMove"
    	}
    	
    	indexHelper = 0;
    	
    	for (Unit unit:unitsMoved)
    	{
    		if (pmUnitMovedWithTransport.get(indexHelper) == null)
    		{
        		List<Unit> currentlyMovedUnit = new ArrayList<Unit>();
				Route route = new Route();
				
				currentlyMovedUnit.add(unit);
				route = pmRoutes.get(indexHelper);

        		//logger.logMessage("" + route + " " + indexHelper);
        		
        		if (route.hasSteps())
        		{
					try
	    			{
	    				originalData.acquireWriteLock();
	    				
	    				moveDel.move(currentlyMovedUnit, route);
	    			}
		    		catch (final Throwable thr)
	    			{
		    			logger.logMessage("Error trying to move with moveDel" + thr);
	    				return;
	    			}
	    			finally
	    			{
	    				originalData.releaseWriteLock();
	    			}
        		}
    		}	
			
    		indexHelper++;
    	}     
        
        //logger.logMessage("delegateMoves: " + moveDel.getMovesMade().toString()); // Part of test log: "performSimulatedMove"
    }	
    
    private HashMap<String, Territory> getTerritoriesToGetInTransports()
    {
    	HashMap<String, Territory> currentTerritoriesToGetInTransports = new LinkedHashMap<String, Territory>();
    	
    	for (TestTransportContainer currentTransport:transports)
    	{
    		for (Entry<String, Territory> territoryWhereTransportCouldBe:currentTransport.getTerritoriesWhereTransportCouldBe().entrySet())
    		{
	    		if (!currentTerritoriesToGetInTransports.containsKey(territoryWhereTransportCouldBe.getKey()))
	    		{
	    			currentTerritoriesToGetInTransports.put(territoryWhereTransportCouldBe.getKey(), territoryWhereTransportCouldBe.getValue());
	    		}
    		}
    	}
    	
    	return currentTerritoriesToGetInTransports;
    }
     
    public double getValueForThisNode(GameData dataOfThisNode, List<String> movedUnitsForEndNode, List<String> unitsForManualInsert, List<Route> routesForManualInsert)
    {
        //logger.logMessage("getGameStatusAtSelectedNode: " + getGameStatusAtSelectedNode(dataOfThisNode));
    	
        double valueOfThisNode = 0;
    	
        TestSetAndEvaluateGoals evaluateGoals = new TestSetAndEvaluateGoals();
        
        // Get and evaluate goals:
        
        // Part of test log: "Main test log"
        /*
        logger.logMessage("-------------------------- MOVE evaluation of goalsList (after rollOut) with size: " + analyseBoard.getOverallStrategyForThisMove().size() 
        		+ " nonCombat: " + nonCombat + " --------------------------");
        */
        
        for (String currentGoal:analyseBoard.getOverallStrategyForThisMove())
        {
        		//logger.logMessage("Test originalData: " + getGameStatusAtSelectedNode(originalData));
        		//logger.logMessage("MOVE getGameStatusThatEvaluationGets: " + getGameStatusAtSelectedNode(dataOfThisNode));	
        		//logger.logMessage("MOVE movedUnitsID Size: " + movedUnitsForEndNode.size()); 
        		//logger.logMessage("depthBeforeRollOut: " + depthBeforeRollOut+1);
        	
        		valueOfThisNode += evaluateGoals.evaluateGoal(currentGoal, dataOfThisNode, playerCopy, movedUnitsForEndNode.subList(0, depthBeforeRollOut+1));
        		
        		//logger.logMessage("MOVE one goal evaluated, valueOfThisNodeNow: " + valueOfThisNode);	// Part of test log: "Main test log"
        }
        
        // Part of test log: "Main test log"
        /*
        logger.logMessage("MOVE all goals evaluated, valueOfThisNode overall: " + valueOfThisNode);
        logger.logMessage(" \n ");
        logger.logMessage(" \n ");
        logger.logMessage(" \n ");
        logger.logMessage(" \n ");
        logger.logMessage(" \n ");
        */
        
    	return valueOfThisNode;
    }
    
    /*
     * ------------------------- Getter -------------------------
     */
    
    public double getTotValue()
    {
    	return totValue;
    }
    
    public double getVisits()
    {
    	return nVisits;
    }
    
    public List<TestUnitAndItsTerritoryContainer> getUnitsThatCanMove()
    {
    	return unitsThatCanMove;
    }
    
    public List<String> getMovedUnits()
    {
    	return movedUnitsID;
    }
    
    public List<Route> getMovedRoutes()
    {
    	return movedRoutes;
    }
    
    /*
     * ------------------------- Methods for testing purposes -------------------------
     */
	 
    /*
     * Returns the stats of this class instance (= the current node) as string
     */
    public String getStats() 
    {
    	String stats = new String();
    	    	
    	stats += " " + nVisits + " " + totValue;

    	return stats;
    }
	
    /*
     * ---- Output of game status ----
     * 
     * Outputs the specified GameData of the map "MiniMap" as formatted string
	 * 
	 * Structure: player name, player territories count, player units count, player factories count, player victory cities count
	 * 
	 * TODO: test and adapt this method, to make it work on all maps
	 */
    public String getGameStatusAtSelectedNode(GameData Data) 
    {    	
    	String getGameStatus = new String();
    	String eol = System.getProperty("line.separator");
    	
    	for (int x = 0; x < Data.getPlayerList().size(); x++)
    	{		
    		int totalUnitCountPlayer = 0;
        	int factories = 0;
        	int victoryCities = 0;
			
			for (Territory t:Data.getMap().getTerritories())
			{	
				if (!t.getName().equals("neutral territory 1") && !t.getName().equals("neutral territory 2") && !t.isWater())
				{	
					totalUnitCountPlayer += t.getUnits().getUnitCount((Data.getPlayerList().getPlayers().get(x))); 
					
					if (t.getUnits().someMatch(new CompositeMatchAnd<Unit>(Matches.unitIsOwnedBy(Data.getPlayerList().getPlayers().get(x)), Matches.UnitCanProduceUnits)))
						factories++;
					
					if (TerritoryAttachment.get(t).getVictoryCity()!=0)
						victoryCities++;
				}
			}
				
			getGameStatus += (eol + "Player name: " + Data.getPlayerList().getPlayers().get(x).toString() 
    				+ " Player territories count: " + Data.getMap().getTerritoriesOwnedBy(Data.getPlayerList().getPlayers().get(x)).size() 
    				+ " Player units count: " + totalUnitCountPlayer + " Player factories count: " + factories + " Player victory cities count: " + victoryCities + eol);
    	}
    	
    	String unitsOnTerritory = new String();
    	ArrayList<UnitType> unitTypeList = new ArrayList<UnitType>();
    	
    	for (Territory t:Data.getMap().getTerritories())
		{
    		for (Unit u:t.getUnits().getUnits())
			{	
				if (!unitTypeList.contains(u.getUnitType()))
					unitTypeList.add(u.getUnitType());
			}		
		}
    	    	
		for (Territory t:Data.getMap().getTerritories())
		{
			if (!t.getName().equals("neutral territory 1") && !t.getName().equals("neutral territory 2") && !t.isWater())
			{	
				unitsOnTerritory += eol + "--------" + t.getName() + "--------" + eol + eol + t.getUnits().getUnitCount() + " Total unit count on this territory (unit name//unit count player 1//unit count player 2)" + eol;
				
				for (int f = 0; f < unitTypeList.size(); f++)
				{	
					// TODO: adapt output to actual player count
					unitsOnTerritory += eol + unitTypeList.get(f).getName() + " " + t.getUnits().getUnitCount(unitTypeList.get(f), Data.getPlayerList().getPlayers().get(0))
										+ " " + t.getUnits().getUnitCount(unitTypeList.get(f), Data.getPlayerList().getPlayers().get(1)) + eol;
				}
		    }
	    }
    	
		getGameStatus += unitsOnTerritory;
		
    	return getGameStatus;
    }
}