package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import java.util.ArrayList;
import java.util.List;

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.NamedAttachable;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.ProductionRule;
import games.strategy.engine.data.Resource;
import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;
import games.strategy.engine.data.UnitType;
import games.strategy.triplea.Constants;
import games.strategy.triplea.attatchments.UnitAttachment;
import games.strategy.triplea.delegate.Matches;
import games.strategy.triplea.delegate.remote.IPurchaseDelegate;
import games.strategy.util.IntegerMap;

/**
 * TestPurchaseAI
 * 
 * @author A.S.
 * @since 2015
 */

public class TestPurchaseAI 
{
	private final static SimpleLogger logger = new SimpleLogger();
	
	private Resource PUs;
	private int leftToSpend;
	private List<ProductionRule> rules;
	private IntegerMap<ProductionRule> purchase;
	private PlayerID player;
	private GameData gameData;
	private ArrayList<TestPurchaseContainer> unitTypesPurchaseList;
	private boolean stopPurchase;
	private boolean harbourForPlacementAvailable;
	
	private double manuallyAddedLandUnitsCount;
	private double manuallyAddedAirUnitsCount;
	private double manuallyAddedSeaUnitsCount;
	private double manuallyAddedAntiAircraftUnitsCount;
	
	private IPurchaseDelegate purchaseDelegate;
	
	private int alignmentOfAI;
	
	public TestPurchaseAI(boolean purchaseForBid, int PUsToSpend, IPurchaseDelegate purchaseDelegate, GameData data, PlayerID player, int alignmentOfAI)
	{			
		this.purchaseDelegate = purchaseDelegate;
		this.alignmentOfAI = alignmentOfAI;
		this.player = player;

		gameData = data;
		
		PUs = data.getResourceList().getResource(Constants.PUS);
		
		leftToSpend = PUsToSpend;
		
		rules = new ArrayList<ProductionRule>();
				
		for (ProductionRule pR:player.getProductionFrontier().getRules())
		{
			rules.add(pR);
		}
		
		unitTypesPurchaseList = new ArrayList<TestPurchaseContainer>();
		
		manuallyAddedLandUnitsCount = 0;
		manuallyAddedAirUnitsCount = 0;
		manuallyAddedSeaUnitsCount = 0;
		manuallyAddedAntiAircraftUnitsCount = 0;
		
		purchase = new IntegerMap<ProductionRule>();
		
		stopPurchase = false;
		
		for (Territory territoryWithOwnFactory:gameData.getMap().getTerritories())
		{			
			if (territoryWithOwnFactory.getOwner().equals(player) && territoryWithOwnFactory.getUnits().someMatch(Matches.UnitCanProduceUnits))
			{
				for (Territory neighbour:gameData.getMap().getNeighbors(territoryWithOwnFactory))
				{
					if (neighbour.isWater())
					{
						harbourForPlacementAvailable = true;
					}
				}
			}
		}	
	}
	
	public void doPurchase() 
	{				
		/*
		 * Step 1: 
		 * 
		 * Create a list with the purchasable unitTypes and their relevant properties for the calculation formula
		 */
		for (ProductionRule pR:rules)
		{
			final NamedAttachable unitNA = pR.getResults().keySet().iterator().next();
			
			if (!(unitNA instanceof UnitType))
			{
				continue; 
			}
			
			final UnitType currentUnitType = (UnitType) unitNA;
			
			// logger.logMessage("unitType: " + currentUnitType.getName() + " getAttack: " + UnitAttachment.get(currentUnitType).getAttack(player));
			
			TestPurchaseContainer currentUnitTypeForPurchase = new TestPurchaseContainer();
			
			currentUnitTypeForPurchase.setUnitTypeName(currentUnitType.getName());
			currentUnitTypeForPurchase.setCostsInPUs(pR.getCosts().getInt(PUs));
			currentUnitTypeForPurchase.setUnitType(currentUnitType);	
			currentUnitTypeForPurchase.setAttack(UnitAttachment.get(currentUnitType).getAttack(player));
			currentUnitTypeForPurchase.setDefense(UnitAttachment.get(currentUnitType).getDefense(player));
			currentUnitTypeForPurchase.setMovement(UnitAttachment.get(currentUnitType).getMovement(player));
			
			unitTypesPurchaseList.add(currentUnitTypeForPurchase);
		}
		
		/* 
		 * Step 2:
		 * 
		 * Calculate the purchase-attractivity of each unit type according to the current alignment of the TestAI player and select the units to be purchased
		 */
		while (stopPurchase == false)
		{
			if (alignmentOfAI == 1) // defensive
			{
				computePurchase(4, 20, 0.5, 3, 1);
			}
			else if (alignmentOfAI == 2) // normal
			{
				computePurchase(2, 30, 1, 1, 1);
			}
			else if (alignmentOfAI == 3) // offensive
			{
				computePurchase(1, 50, 3, 0.5, 1);
			}
			else
			{
				logger.logMessage("Purchase stopped: no valid alignment of TestAI player");
			}
		}
		
		/* 
		 * Step 3: 
		 * 
		 * Purchase the units that were chosen in computePurchase() with the corresponding delegate
		 */
		// logger.logMessage("Purchase: " + purchase.keySet().iterator().next());
		// logger.logMessage("" + purchaseDelegate.purchase(purchase));
		purchaseDelegate.purchase(purchase);
	}
	
	/*
	 * Evaluates the purchase-attractivity of the available unit types and selects a single unit to be purchased
	 */
	private void computePurchase(int minimumPUsLeft, double costWeight, double attackWeight, double defenseWeight, double movementWeight)
	{
		/*
		 * Step 1:
		 * 
		 * Calculate the unit- and territory-ratios to be used as multiplier in the formulas of step 2 then
		 */
		 
		// Count units and territories
		// TODO: would be more efficient to do this once in doPurchase() before while-loop?
		double ownLandUnitsCounter = 0;
		double ownAirUnitsCounter = 0;
		double ownSeaUnitsCounter = 0;
		double ownAntiAircrafUnitsCounter = 0;
		double ownTerritoriesCounter = 0;
		
		double enemyLandUnitsCounter = 0;
		double enemyAirUnitsCounter = 0;
		double enemySeaUnitsCounter = 0;
		double enemyAntiAircrafUnitsCounter = 0;
		double enemyTerritoriesCounter = 0;
		
		for (Territory currentTerritory:gameData.getMap().getTerritories())
		{
			if (currentTerritory.getOwner().equals(player))
			{
				ownTerritoriesCounter++;
			}
			else
			{
				enemyTerritoriesCounter++;
			}
			
			for (Unit currentUnit:currentTerritory.getUnits())
			{
				if (currentUnit.getOwner().equals(player))
				{
					if (!(Matches.UnitIsAir.match(currentUnit) && Matches.UnitIsSea.equals(currentUnit)))
					{
						ownLandUnitsCounter++;
					}
					
					if (Matches.UnitIsAir.match(currentUnit))
					{
						ownAirUnitsCounter++;
					}
					
					if (Matches.UnitIsSea.match(currentUnit))
					{
						ownSeaUnitsCounter++;
					}
					
					if (UnitAttachment.get(currentUnit.getUnitType()).getIsAAforCombatOnly())
					{
						ownAntiAircrafUnitsCounter++;
					}
				}
				else if (!currentUnit.getOwner().equals(player))
				{
					if (!(Matches.UnitIsAir.match(currentUnit) && Matches.UnitIsSea.equals(currentUnit)))
					{
						enemyLandUnitsCounter++;
					}
					
					if (Matches.UnitIsAir.match(currentUnit))
					{
						enemyAirUnitsCounter++;
					}
					
					if (Matches.UnitIsSea.match(currentUnit))
					{
						enemySeaUnitsCounter++;
					}
					
					if (UnitAttachment.get(currentUnit.getUnitType()).getIsAAforCombatOnly())
					{
						enemyAntiAircrafUnitsCounter++;
					}
				}
				else
				{
					logger.logMessage("Unit does not belong to any player");
				}
			}
		}
		
		// Calculate unit- and territory-ratios
		double landUnitsRatio = 0;
		double airUnitsRatio = 0;
		double seaUnitsRatio = 0;
		double ownAntiAircraftToEnemyAirUnitsRatio = 0;
		double territoriesRatio = 0;
		
		if (ownLandUnitsCounter != 0 && enemyLandUnitsCounter != 0)
		{
			landUnitsRatio = (ownLandUnitsCounter+manuallyAddedLandUnitsCount)/enemyLandUnitsCounter;
		}
		else
		{
			landUnitsRatio = 1; // TODO: set an appropriate value for this case
		}
		
		if (ownAirUnitsCounter != 0 && enemyAirUnitsCounter != 0)
		{
			airUnitsRatio = (ownAirUnitsCounter+manuallyAddedAirUnitsCount)/enemyAirUnitsCounter;
		}
		else
		{
			airUnitsRatio = 1; // TODO: set an appropriate value for this case
		}
		
		if (ownSeaUnitsCounter != 0 && enemySeaUnitsCounter != 0)
		{
			seaUnitsRatio = (ownSeaUnitsCounter+manuallyAddedSeaUnitsCount)/enemySeaUnitsCounter;
		}
		else
		{
			seaUnitsRatio = 1; // TODO: set an appropriate value for this case
		}
		
		if (ownAntiAircrafUnitsCounter != 0 && enemyAirUnitsCounter != 0)
		{
			ownAntiAircraftToEnemyAirUnitsRatio = (ownAntiAircrafUnitsCounter+manuallyAddedAntiAircraftUnitsCount)/enemyAirUnitsCounter;
		}
		else
		{
			ownAntiAircraftToEnemyAirUnitsRatio = 1; // TODO: set an appropriate value for this case
		}
		
		if (ownTerritoriesCounter != 0 && enemyTerritoriesCounter != 0)
		{
			territoriesRatio = ownTerritoriesCounter/enemyTerritoriesCounter;
		}
		else
		{
			territoriesRatio = 1; // TODO: set an appropriate value for this case
		}
		
		/*
		 * Step 2:
		 * 
		 * Calculate the purchase-attractivity for each available unit type based on the current unit- and territory-ratios
		 */
		// logger.logMessage("---------------- Start run for evaluation of purchase-attractivity ----------------");
		for (int i = 0; i < rules.size(); i++)
		{
			/*
			 * Note: check, if it is a seaUnit, and we do (not) have appropriate factories for their placement
			 */
			if (!(Matches.UnitTypeIsSea.match(unitTypesPurchaseList.get(i).getUnitType()) && !harbourForPlacementAvailable))
			{				
				// logger.logMessage("---------------- Current unit: " + unitTypesPurchaseList.get(i).getUnitType().getName() + " ----------------");
				
				// Can we afford the unit?
				if (unitTypesPurchaseList.get(i).getCostsInPUs() <= (leftToSpend-minimumPUsLeft))
				{	
					double purchaseAttractivity = 0;
						
					// 2.1: Calculate purchaseAttractivity: basic properties of unit
					purchaseAttractivity = (costWeight-unitTypesPurchaseList.get(i).getCostsInPUs())* // unit costs
							(unitTypesPurchaseList.get(i).getAttack()*attackWeight+ // attack * a
							unitTypesPurchaseList.get(i).getDefense()*defenseWeight+ // defensve * b
							unitTypesPurchaseList.get(i).getMovement()*movementWeight); // movement * c
					/*
					logger.logMessage("Cost value of the formula: " + costWeight + " - " + unitTypesPurchaseList.get(i).getCostsInPUs() + " = " + (costWeight-unitTypesPurchaseList.get(i).getCostsInPUs()));
					logger.logMessage("Attack value of the formula: " + unitTypesPurchaseList.get(i).getAttack() + " * " + attackWeight + " = " + unitTypesPurchaseList.get(i).getAttack()*attackWeight);
					logger.logMessage("Defense value of the formula: " + unitTypesPurchaseList.get(i).getDefense() + " * " + defenseWeight + " = " + unitTypesPurchaseList.get(i).getDefense()*defenseWeight);
					logger.logMessage("Movement value of the formula: " + unitTypesPurchaseList.get(i).getMovement() + " * " + movementWeight + " = " + unitTypesPurchaseList.get(i).getMovement()*movementWeight);
					
					logger.logMessage("purchaseAttractivity after basic properties of unit type: " + (costWeight-unitTypesPurchaseList.get(i).getCostsInPUs()) + " * "
					+ (unitTypesPurchaseList.get(i).getAttack()*attackWeight+unitTypesPurchaseList.get(i).getDefense()*defenseWeight+unitTypesPurchaseList.get(i).getMovement()*movementWeight) + " = " + purchaseAttractivity);
					*/
					
					// 2.2: Calculate purchaseAttractivity: special properties of unit						
					
					// If the TestAI player holds few territories, and has no defensive alignment, "blitzing" unit types shall be more attractive 
					if (UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getCanBlitz(player) && alignmentOfAI != 1)
					{
						// logger.logMessage("Blitz value of the formula: " + (purchaseAttractivity * (territoriesRatio*10)));
						purchaseAttractivity = purchaseAttractivity * (territoriesRatio*10); 	
					}
					
					// If the enemy players have many air units, and the alignment of the TestAI player is not offensive, specialized anti-aircraft unit types shall be more attractive
					if (UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getIsAAforCombatOnly() && alignmentOfAI != 3)
					{
						//logger.logMessage("AntiAircraft value of the formula: " + (purchaseAttractivity*(ownAntiAircraftToEnemyAirUnitsRatio*10)));
						purchaseAttractivity = purchaseAttractivity * (ownAntiAircraftToEnemyAirUnitsRatio*10); 		
					}
					
					//logger.logMessage("purchaseAttractivity after special properties of unit type (as multiplier): " + purchaseAttractivity);
					
					// 2.3: Calculate purchaseAttractivity: unitType of unit
					if (!UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getIsAir() && UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getIsMarine() == 0)
					{
						//logger.logMessage("UnitType is land value of the formula: " + purchaseAttractivity/landUnitsRatio);
						purchaseAttractivity = purchaseAttractivity/landUnitsRatio;
					}
					else if (UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getIsAir())
					{
						//logger.logMessage("UnitType is air value of the formula: " + purchaseAttractivity/airUnitsRatio);
						purchaseAttractivity = purchaseAttractivity/airUnitsRatio;
					}
					else if (UnitAttachment.get(unitTypesPurchaseList.get(i).getUnitType()).getIsMarine() == 1)
					{
						//logger.logMessage("UnitType is marine value of the formula: " + purchaseAttractivity/seaUnitsRatio);
						purchaseAttractivity = purchaseAttractivity/seaUnitsRatio;
					}
					else
					{
						logger.logMessage("Unit does not belong to any player");
					}
					
					//logger.logMessage("purchaseAttractivity after unit type (as multiplier): " + purchaseAttractivity);
					
					// 2.4: Calculate purchaseAttractivity: own unit type ratio artillery/infantry
					// TODO: could make sense to balance ratio between own artillery and infantry
					
					//logger.logMessage("purchaseAttractivity after own unit type ratio artillery/infantry: " + purchaseAttractivity);
					
					unitTypesPurchaseList.get(i).setPurchaseAttractivity(purchaseAttractivity);
				}	
			}
			else 
			{
				continue;
			}
		}
		// logger.logMessage("---------------- End of run for evaluation of purchase-attractivity ----------------");
		
		/*
		 * Step 3:
		 *
		 * Select the unit type with the highest purchase-attractivity, that is still affordable
		 *
		 * TODO: should be possible to integrate this in step 2
		 */
		double dummyPurchaseAttractivity = 0;
		int dummyIndex = 0;
		int cheapestUnitTypeCosts = Integer.MAX_VALUE;

		for (int i = 0; i < rules.size(); i++)
		{
			//logger.logMessage("Name: " + unitTypesPurchaseList.get(i).getUnitTypeName() + " purchase-attractivity " + unitTypesPurchaseList.get(i).getPurchaseAttractivity());
			
			// purchaseAttractivity = 0 (default value), if unit type could not be afforded in the evaluation in step 2
			if (unitTypesPurchaseList.get(i).getPurchaseAttractivity() != 0)
			{
				if (unitTypesPurchaseList.get(i).getPurchaseAttractivity() > dummyPurchaseAttractivity && unitTypesPurchaseList.get(i).getCostsInPUs() <= (leftToSpend-minimumPUsLeft)) 
				{
					//logger.logMessage("Name: " + unitTypesPurchaseList[i][0] + " purchase-attractivity " + unitTypesPurchaseList[i][7]);
					dummyPurchaseAttractivity = unitTypesPurchaseList.get(i).getPurchaseAttractivity();
					dummyIndex = i;
				}
				
				// get costs of cheapest unit type for the calculation of stopPurchase at the end
				if (unitTypesPurchaseList.get(i).getCostsInPUs() < cheapestUnitTypeCosts)
				{
					cheapestUnitTypeCosts = unitTypesPurchaseList.get(i).getCostsInPUs();
				}
			}	
		}
		
		//logger.logMessage("Name of the chosen unit type in this run: " + unitTypesPurchaseList.get(dummyIndex).getUnitTypeName() + " purchase-attractivity " + unitTypesPurchaseList.get(dummyIndex).getPurchaseAttractivity());
		
		/*
		 * Step 4:
		 *
		 * Add the chosen unit type to the purchase list
		 */
		if (unitTypesPurchaseList.get(dummyIndex).getCostsInPUs() <= (leftToSpend-minimumPUsLeft))
		{
			if (!Matches.UnitTypeIsAir.match(unitTypesPurchaseList.get(dummyIndex).getUnitType()) && !Matches.UnitTypeIsSea.match(unitTypesPurchaseList.get(dummyIndex).getUnitType()))
			{
				manuallyAddedLandUnitsCount++;
			}
			else if (Matches.UnitTypeIsAir.match(unitTypesPurchaseList.get(dummyIndex).getUnitType()))
			{
				manuallyAddedAirUnitsCount++;
			}
			else if (Matches.UnitTypeIsSea.match(unitTypesPurchaseList.get(dummyIndex).getUnitType()))
			{
				manuallyAddedSeaUnitsCount++;
			}
			else if (Matches.UnitTypeIsAAforCombatOnly.match(unitTypesPurchaseList.get(dummyIndex).getUnitType()))
			{
				manuallyAddedAntiAircraftUnitsCount++;
			}
			else
			{
				logger.logMessage("invalid unit type");
			}
			
			leftToSpend -= unitTypesPurchaseList.get(dummyIndex).getCostsInPUs();
			// logger.logMessage("leftToSpend: " + leftToSpend);
			
			// Unit will be purchased -				
			purchase.add(rules.get(dummyIndex), 1);	
		}	
		else if (minimumPUsLeft > (leftToSpend-cheapestUnitTypeCosts))
		{	
			stopPurchase = true;
			// logger.logMessage("Abort calculation of unit to be purchased: not enough PUs left to purchase any unit type");
		}
	}
}