package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.Route;
import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;
import games.strategy.triplea.attatchments.TerritoryAttachment;
import games.strategy.triplea.attatchments.UnitAttachment;
import games.strategy.triplea.delegate.Matches;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

/**
 * TestSetAndEaluateGoals
 * 
 * Used for defining and evaluating goals for the move-phase
 * 
 * @author A.S.
 * @since 2015
 */

public class TestSetAndEvaluateGoals 
{
	private final static SimpleLogger logger = new SimpleLogger();
	
	private HashMap<String, Territory> relevantTerritoriesForGoal;

	private double valueOfThisGoal;
		
	private int ownTerritoryDistance;
	private int enemyTerritoryDistance;
	
	private double ownUnitsCounter;
	private double enemyUnitsCounter;
	
	// Number of (expectedly) won territories of the move
	private double territoriesWon;
	// Number of (expectedly) lost units of the move
	private double ownUnitsLost;
	// Number of territories with at least one own unit
	private double territoriesOccupied;
	
	public TestSetAndEvaluateGoals()
	{
		ownTerritoryDistance = 0;
		enemyTerritoryDistance = 0;
		
		ownUnitsCounter = 0;
		enemyUnitsCounter = 0;
		territoriesWon = 0;
	}
		
	public HashMap<String, Territory> setRelevantTerritoriesForGoals(ArrayList<String> goalsList, GameData gameData, PlayerID player)
	{
		relevantTerritoriesForGoal = new HashMap<String, Territory>();
		
		for (String currentGoal:goalsList)
		{
			if (currentGoal.equals("More own units than enemy units in area around own territory"))
			{
				if (setTerritoriesForMoreOwnUnitsThanEnemyUnits(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForMoreOwnUnitsThanEnemyUnits(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
			else if (currentGoal.equals("At least one own unit on own territory"))
			{
				if (setTerritoriesForAtLeastOneOwnUnit(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForAtLeastOneOwnUnit(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
			else if (currentGoal.equals("Build and supply frontline"))
			{	
				if (setTerritoriesForBuildAndSupplyFrontline(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForBuildAndSupplyFrontline(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
			else if (currentGoal.equals("Controlled attack of nearest enemy territories"))
			{	
				if (setTerritoriesForControlledAttack(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForControlledAttack(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
			else if (currentGoal.equals("Build defense for own victory cities"))
			{
				if (setTerritoriesForBuildDefenseForOwnVictoryCities(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForBuildDefenseForOwnVictoryCities(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
			else if (currentGoal.equals("Defend marine landing points"))
			{
				if (setTerritoriesForDefendMarineLandingPoints(gameData, player).size() != 0)
				{
					for (Entry<String, Territory> currentTerritory:setTerritoriesForDefendMarineLandingPoints(gameData, player).entrySet())
					{
						relevantTerritoriesForGoal.put(currentTerritory.getKey(), currentTerritory.getValue());
					}
				}
			}
		}	
		
		return relevantTerritoriesForGoal;
	}
	
	public double evaluateGoal(String currentGoal, GameData gameData, PlayerID player, List<String> movedUnitsID)
	{
		if (currentGoal.equals("More own units than enemy units in area around own territory"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateMoreOwnUnitsThanEnemyUnits(gameData, player, movedUnitsID);
		}
		else if (currentGoal.equals("At least one own unit on own territory"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateAtLeastOneOwnUnit(gameData, player, movedUnitsID);
		}
		else if (currentGoal.equals("Build and supply frontline"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateBuildAndSupplyFrontline(gameData, player, movedUnitsID);
		}
		else if (currentGoal.equals("Controlled attack of nearest enemy territories"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateControlledAttack(gameData, player, movedUnitsID);
		}
		else if (currentGoal.equals("Build defense for own victory cities"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateBuildDefenseForOwnVictoryCities(gameData, player, movedUnitsID);
		}
		else if (currentGoal.equals("Defend marine landing points"))
		{
			// logger.logMessage("-------- SAEG evaluation of currentGoal: " + currentGoal + " --------"); // Part of test log: "Main test log"
			valueOfThisGoal = evaluateDefendMarineLandingPoints(gameData, player, movedUnitsID);
		}
		
		return valueOfThisGoal;
	}
	
	/*
	 * TODO: overall, the different evaluation formulas should follow a common concept in order to be (reasonably) combinable and comparable with each other
	 */
	
	private HashMap<String, Territory> setTerritoriesForMoreOwnUnitsThanEnemyUnits(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (t.getOwner().equals(player))
			{
				territoriesToMoveToForThisGoal.put(t.getName().toString(), t);
			}
		}
		
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateMoreOwnUnitsThanEnemyUnits(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		territoriesWon = 1;
		
		ownTerritoryDistance = 1;
		enemyTerritoryDistance = 1;
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			ownUnitsCounter = 0;
			enemyUnitsCounter = 0;

			if (t.getOwner().equals(player))
			{
				// logger.logMessage("SAEG Territory: " + t.getName()); // Part of test log: "Main test log"
				
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= ownTerritoryDistance && Matches.isTerritoryAllied(player, gameData).match(territoryWithinDistanceX))
					{
						// TODO: check, if allied units are considered here too
						ownUnitsCounter += territoryWithinDistanceX.getUnits().getUnitCount();
					}
					else if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= enemyTerritoryDistance && Matches.isTerritoryEnemy(player, gameData).match(territoryWithinDistanceX))
					{
						// TODO: don't count neutral units here
						enemyUnitsCounter += territoryWithinDistanceX.getUnits().getUnitCount();
					}
				}
				// logger.logMessage("SAEG ownUnitsCounter: " + ownUnitsCounter + " enemyUnitsCounter: " + enemyUnitsCounter); // Part of test log: "Main test log"
				
	    		if (ownUnitsCounter > enemyUnitsCounter)
	    		{
	    			territoriesWon += 0.3;
	    			
	    			valueOfThisGoal *= 1;
					// logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " + 1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
	    		}
	    		else
	    		{
	    			valueOfThisGoal *= (1-ownUnitsCounter/10);
	    			// logger.logMessage("SAEG Goal not accomplished for this territory -> valueOfThisGoal * " + (1-ownUnitsCounter/10) + " = " + valueOfThisGoal); // Part of test log: "Main test log"
	    		}
			}	
		}
 	
		valueOfThisGoal *= territoriesWon;
		// logger.logMessage("SAEG Multiply value of goal with the number of 'won' territories -> valueOfThisGoal * " + territoriesWon + " = " + valueOfThisGoal); // Part of test log: "Main test log"
		
		return valueOfThisGoal;
	}
	
	private HashMap<String, Territory> setTerritoriesForAtLeastOneOwnUnit(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (t.getOwner().equals(player))
			{
				territoriesToMoveToForThisGoal.put(t.getName().toString(), t);
			}
		}
		
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateAtLeastOneOwnUnit(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		territoriesOccupied = 1;
		
		for (Territory t:gameData.getMap().getTerritories())
		{
			if (t.getOwner().equals(player))
			{
				// logger.logMessage("SAEG Territory: " + t.getName()); // Part of test log: "Main test log"
				
				if (t.getUnits().getUnitCount() > 0)
				{
					territoriesOccupied += 0.3;
					
					valueOfThisGoal *= 1;
					
					// logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " + 1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
				}
				else
				{
					// valueOfThisGoal -= 1;
					
					// logger.logMessage("SAEG Goal not accomplished for this territory -> valueOfThisGoal * " + -1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
				}
			}	
		}
		
		valueOfThisGoal *= territoriesOccupied;
		
		return valueOfThisGoal;
	}
	
	private HashMap<String, Territory> setTerritoriesForBuildAndSupplyFrontline(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (t.getOwner().equals(player))
			{
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{
					ownUnitsCounter = 0;
					
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= 4
							&& Matches.isTerritoryEnemyAndNotUnownedWaterOrImpassibleOrRestricted(player, gameData).match(territoryWithinDistanceX)
							&& !territoriesToMoveToForThisGoal.containsKey(territoryWithinDistanceX.getName().toString()))
					{
						territoriesToMoveToForThisGoal.put(t.getName().toString(), t);
					}
				}
			}
		}
		
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateBuildAndSupplyFrontline(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		territoriesWon = 1;
		
		for (Territory t:gameData.getMap().getTerritories())
		{				
			if (t.getOwner().equals(player))
			{
				// logger.logMessage("SAEG Territory for possible attack: " + t.getName()); // Part of test log: "Main test log"
				
				/*
				 * TODO: check plausibility of this condition 
				 * 
				 * -> if we use BuildAndSupplyFrontline after combatMove(), should the won territories be occupied with own units?
				 */
				/*
	    		if (t.getUnits().getUnitCount() - t.getUnits().getUnitCount(player) == 0)
	    		{	    				
	    			valueOfThisGoal += 1;
	    			logger.logMessage("SAEG Goal accomplished for this territory -> +1");
	    		}
				*/
				
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{
					ownUnitsCounter = 0;
					
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= 4 && Matches.isTerritoryEnemyAndNotUnownedWaterOrImpassibleOrRestricted(player, gameData).match(territoryWithinDistanceX))
					{
						// logger.logMessage("--- SAEG adjacent own Territory: " + territoryWithinDistanceX.getName() + " ---"); // Part of test log: "Main test log"
						
						ownUnitsCounter += t.getUnits().getUnitCount();
						
						// logger.logMessage("---- SAEG ownUnitsCounter on own territory to possibly attack from: " + ownUnitsCounter + " enemyUnitsCounter on territory for possible attack: " + t.getUnits().getUnitCount() + " ----"); // Part of test log: "Main test log"
						
						// front line 
						if (gameData.getMap().getDistance(t, territoryWithinDistanceX) == 1)
						{
							// TODO: make sure, that the possible battle would be won
				    		if (ownUnitsCounter > territoryWithinDistanceX.getUnits().getUnitCount() + 1)	
				    		{	    	
				    			territoriesWon += t.getUnits().getUnitCount()*(1-gameData.getMap().getDistance(t, territoryWithinDistanceX)*0.1);
				    			
				    			valueOfThisGoal *= 1;
				    			
				    			// logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " + 1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
				    		}
				    		else
				    		{ 
				    			valueOfThisGoal *= (1-ownUnitsCounter/10); 
				    			
				    			// logger.logMessage("SAEG Unit would be lost -> valueOfThisGoal * " + (1-ownUnitsCounter/10) + " = " + valueOfThisGoal); // Part of test log: "Main test log"
				    		}	
						}	
						else
						{
							// "the more own units near the front line, the better"
							territoriesWon += t.getUnits().getUnitCount()*(1-gameData.getMap().getDistance(t, territoryWithinDistanceX)*0.1);
						}
					}
				}
			}	
		}
 	
		valueOfThisGoal *= territoriesWon;
		// logger.logMessage("SAEG Multiply value of goal with the number of 'won' territories -> valueOfThisGoal * " + territoriesWon + " = " + valueOfThisGoal); // Part of test log: "Main test log"
		
		return valueOfThisGoal;
	}
	
	private HashMap<String, Territory> setTerritoriesForControlledAttack(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (Matches.isTerritoryEnemyAndNotUnownedWaterOrImpassibleOrRestricted(player, gameData).match(t))
			{
				territoriesToMoveToForThisGoal.put(t.getName().toString(), t);
			}
		}	
		/*
		logger.logMessage("SAEG territoriesToMoveToForThisGoal size: " + territoriesToMoveToForThisGoal.size());
		
		for (String t:territoriesToMoveToForThisGoal)
		{
			logger.logMessage("SAEG territoriesToMoveToForThisGoal: " + t);
		}
		*/
		
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateControlledAttack(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		territoriesWon = 1;
		ownUnitsLost = 0;
		
		/*
		logger.logMessage("size: " + movedUnitsID.size()); // Part of test log: "Main test log"
		
		for (String s:movedUnitsID)
		{
			logger.logMessage("already moved unit: " + s); // Part of test log: "Main test log"
		}
		*/
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			ownUnitsCounter = 0;
			enemyUnitsCounter = 0;
			
			if (Matches.isTerritoryEnemyAndNotUnownedWaterOrImpassibleOrRestricted(player, gameData).match(t) && t.getUnits().getUnitCount(player) > 0)
			{	
				ownUnitsCounter = t.getUnits().getUnitCount(player);
				enemyUnitsCounter = (t.getUnits().getUnitCount()-t.getUnits().getUnitCount(player)); // TODO: there should be a cleaner way to get the count of enemy units on a territory 
				
				/*
				// TODO: should we consider a counter-attack here too?
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= ownTerritoryDistance && Matches.isTerritoryAllied(player, gameData).match(territoryWithinDistanceX))
					{
						ownUnitsCounter += territoryWithinDistanceX.getUnits().getUnitCount();
					}
					else if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= enemyTerritoryDistance && Matches.isTerritoryEnemy(player, gameData).match(territoryWithinDistanceX))
					{
						enemyUnitsCounter += territoryWithinDistanceX.getUnits().getUnitCount();
					}
				}
				*/
				
				// logger.logMessage("SAEG Territory to attack: " + t.getName() + " ownUnitsCounter: " + ownUnitsCounter + " enemyUnitsCounter: " + enemyUnitsCounter); // Part of test log: "Main test log"
				
				// TODO: make sure, that the possible battle would be won
	    		if (ownUnitsCounter > enemyUnitsCounter + 1)
	    		{	    				
	    			territoriesWon += 1;
	    			
	    			// logger.logMessage("SAEG Goal accomplished for this territory -> territoriesWon +" + 1); // Part of test log: "Main test log"
	    		}
	    		else
	    		{ 
	    			ownUnitsLost += ownUnitsCounter; 
	    			
	    			// logger.logMessage("SAEG Goal not accomplished for this territory: " + ownUnitsCounter + " own unit(s) would probably be lost -> ownUnitsLost +" + ownUnitsCounter); // Part of test log: "Main test log"
	    		}
			}	
		}
		
		valueOfThisGoal *= territoriesWon-(ownUnitsLost*0.5);
		
		// logger.logMessage("SAEG Multiply value of goal with the number of 'won' territories minus own units lost -> valueOfThisGoal * " + (territoriesWon-(ownUnitsLost*0.5)) + " = " + valueOfThisGoal); // Part of test log: "Main test log"
		
		return valueOfThisGoal;
	}
	
	private HashMap<String, Territory> setTerritoriesForBuildDefenseForOwnVictoryCities(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (Matches.isTerritoryOwnedBy(player).match(t) && TerritoryAttachment.get(t).getVictoryCity() == 1)
			{
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{					
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= 2 && Matches.isTerritoryAllied(player, gameData).match(territoryWithinDistanceX)
							&& !territoriesToMoveToForThisGoal.containsKey(territoryWithinDistanceX.getName().toString()))
					{
						territoriesToMoveToForThisGoal.put(territoryWithinDistanceX.getName().toString(), territoryWithinDistanceX);
					}
				}
			}
		}	
		/*
		logger.logMessage("SAEG territoriesToMoveToForThisGoal size: " + territoriesToMoveToForThisGoal.size());
		
		for(String t:territoriesToMoveToForThisGoal)
		{
			logger.logMessage("SAEG territoriesToMoveTo: " + t);
		}
		*/ 
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateBuildDefenseForOwnVictoryCities(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		
		for (Territory t:gameData.getMap().getTerritories())
		{			
			if (Matches.isTerritoryOwnedBy(player).match(t) && TerritoryAttachment.get(t).getVictoryCity() == 1)
			{
				for (Territory territoryWithinDistanceX:gameData.getMap().getTerritories())
				{					
					if (gameData.getMap().getDistance(t, territoryWithinDistanceX) <= 2 && Matches.isTerritoryAllied(player, gameData).match(territoryWithinDistanceX))
					{
						if (t.getUnits().getUnitCount(player) > 0)
						{	
							if (gameData.getMap().getDistance(t, territoryWithinDistanceX) > 0)
							{	
								valueOfThisGoal *= (1+t.getUnits().getUnitCount(player)/10) / gameData.getMap().getDistance(t, territoryWithinDistanceX);
								
								/*
								logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " 
										+ (1+t.getUnits().getUnitCount(player)/10) / gameData.getMap().getDistance(t, territoryWithinDistanceX) 
										+ " = " + valueOfThisGoal); // Part of test log: "Main test log"
								*/
							}
							else
							{
								valueOfThisGoal *= (1+t.getUnits().getUnitCount(player)/10);
								
								/*
								logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " 
										+ (1+t.getUnits().getUnitCount(player)/10) + " = " + valueOfThisGoal); // Part of test log: "Main test log"
								*/
							}		
						}
						else
						{
							valueOfThisGoal *= -1;
							//logger.logMessage("SAEG Goal not accomplished for this territory -> valueOfThisGoal * " + -1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
						}
					}
				}
			}
		}
		
		return valueOfThisGoal;
	}
	
	private HashMap<String, Territory> setTerritoriesForDefendMarineLandingPoints(GameData gameData, PlayerID player)
	{
		HashMap<String, Territory> territoriesToMoveToForThisGoal = new HashMap<String, Territory>();
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			if (Matches.isTerritoryOwnedBy(player).match(t) && Matches.territoryHasNeighborMatching(gameData, Matches.TerritoryIsWater).match(t))
			{
				/*
				 * -> land territory should be defended too, e.g., if enemy transport is arriving with several units
				 */
				if (!territoriesToMoveToForThisGoal.containsKey(t.getName().toString()))
				{
					territoriesToMoveToForThisGoal.put(t.getName().toString(), t);
				}
				
				for (Territory neighbour:gameData.getMap().getNeighbors(t))
				{
					if (neighbour.isWater() && !territoriesToMoveToForThisGoal.containsKey(neighbour.getName().toString()))
					{
						territoriesToMoveToForThisGoal.put(neighbour.getName().toString(), neighbour);
					}
				}
			}
		}	
		/*
		logger.logMessage("SAEG territoriesToMoveToForThisGoal size: " + territoriesToMoveToForThisGoal.size());
		
		for (String t:territoriesToMoveToForThisGoal)
		{
			logger.logMessage("SAEG territoriesToMoveTo: " + t);
		}
		*/
		
		return territoriesToMoveToForThisGoal;
	}
	
	private double evaluateDefendMarineLandingPoints(GameData gameData, PlayerID player, List<String> movedUnitsID) 
	{
		valueOfThisGoal = 1;
		territoriesWon = 1;
		
		/*
		// Test
		logger.logMessage("size: " + movedUnitsID.size());
		
		for (String s:movedUnitsID)
		{
			logger.logMessage("already moved unit: " + s);
		}
		*/
		
		for (Territory t:gameData.getMap().getTerritories())
		{	
			ownUnitsCounter = 0;
			enemyUnitsCounter = 0;
			
			if (Matches.isTerritoryOwnedBy(player).match(t) && Matches.territoryHasNeighborMatching(gameData, Matches.TerritoryIsWater).match(t))
			{
				// TODO: we should not only consider neighbour-territories, because enemy transports could load and unload from territories further away in the next move, and therefore are already a threat
				for (Territory neighbour:gameData.getMap().getNeighbors(t))
				{
					if (neighbour.isWater())
					{						
						ownUnitsCounter = t.getUnits().getUnitCount(player) + neighbour.getUnits().getUnitCount(player);
						enemyUnitsCounter = (neighbour.getUnits().getUnitCount() - neighbour.getUnits().getUnitCount(player));
					}
				}
				
				// logger.logMessage("SAEG Territory to defend: " + t.getName() + " ownUnitsCounter: " + ownUnitsCounter + " enemyUnitsCounter: " + enemyUnitsCounter); // Part of test log: "Main test log"
				
				// TODO: make sure, that the territory can be defended on longer term (if it is strategically important)
	    		if (ownUnitsCounter > enemyUnitsCounter || enemyUnitsCounter == 0)
	    		{	    				
	    			territoriesWon += 1;
	    			
	    			valueOfThisGoal *= 1;
	    			
	    			// logger.logMessage("SAEG Goal accomplished for this territory -> valueOfThisGoal * " + 1 + " = " + valueOfThisGoal); // Part of test log: "Main test log"
	    		}
	    		else
	    		{ 
	    			valueOfThisGoal *= (1-ownUnitsCounter/10); 
	    			
	    			// logger.logMessage("SAEG Unit would be lost -> valueOfThisGoal * " + (1-ownUnitsCounter/10) + " = " + valueOfThisGoal); // Part of test log: "Main test log"
	    		}
			}		
		}
 	
		valueOfThisGoal *= territoriesWon;
	
		// logger.logMessage("SAEG Multiply value of goal with the number of 'won' territories -> valueOfThisGoal * " + territoriesWon + " = " + valueOfThisGoal); // Part of test log: "Main test log"
		
		return valueOfThisGoal;
	}
}