package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

/**
 * TestTransportContainer
 *
 * A container for storing all kinds of information for units of the unitType 'transport'
 * 
 * @author A.S.
 * @since 2015
 */
public class TestTransportContainer 
{
	private Unit unitObjectOfTransport;
		
	private ArrayList<Unit> loadedUnits;
	
	private int loadedUnitsCosts;
	
	private HashMap<String, Territory> territoriesWhereTransportCouldBe;
	
	private HashMap<String, Territory> territoriesWhereTransportCanUnload;
	
	private HashMap<String, Territory> territoriesWhereTransportCanUnloadCopy;
	
	private boolean canTransportMove;
	
	private List<Territory> currentRoute;
	
	public TestTransportContainer(Unit unitObjectOfTransport)
	{	
		this.unitObjectOfTransport = unitObjectOfTransport;
		
		loadedUnits = new ArrayList<Unit>();
		
		loadedUnitsCosts = 0;
		
		territoriesWhereTransportCouldBe = new HashMap<String, Territory>();
		
		territoriesWhereTransportCanUnload = new HashMap<String, Territory>();
		
		canTransportMove = true;
	}
	
	public Unit getUnitObjectOfTransport()
	{
		return unitObjectOfTransport;
	}
	
	public void loadUnit(Unit unitToLoad)
	{
		loadedUnits.add(unitToLoad);
	}
	
	public void unloadUnit(Unit unitToUnload)
	{
		loadedUnits.remove(unitToUnload);
	}
	
	public ArrayList<Unit> getLoadedUnits()
	{
		return loadedUnits;
	}
	
	public void changeLoadedUnitsCosts(int loadedUnitsCosts)
	{
		this.loadedUnitsCosts += loadedUnitsCosts;
	}
	
	public int getLoadedUnitsCosts()
	{
		return loadedUnitsCosts;
	}
	
	public void addTerritoryWhereTransportCouldBe(Territory territoryWhereTransportCouldBe)
	{
		territoriesWhereTransportCouldBe.put(territoryWhereTransportCouldBe.getName().toString(), territoryWhereTransportCouldBe);
	}
	
	public void clearTerritoriesWhereTransportCouldBe()
	{
		territoriesWhereTransportCouldBe.clear();
	}
	
	public HashMap<String, Territory> getTerritoriesWhereTransportCouldBe()
	{		
		return territoriesWhereTransportCouldBe;
	}
	
	public void addTerritoryWhereTransportCanUnload(Territory territoryWhereTransportCanUnload)
	{
		territoriesWhereTransportCanUnload.put(territoryWhereTransportCanUnload.getName().toString(), territoryWhereTransportCanUnload);
	}
	
	public void clearTerritoriesWhereTransportCanUnload()
	{
		territoriesWhereTransportCanUnload.clear();
	}
	
	public HashMap<String, Territory> getTerritoriesWhereTransportCanUnload()
	{
		return territoriesWhereTransportCanUnload;
	}
	
	/*
	 * Deep copy
	 */
	public void copyTerritoriesWhereTransportCanUnload()
	{
		territoriesWhereTransportCanUnloadCopy = new HashMap<String, Territory>();
		
		for (Entry<String, Territory> currentTerritory:this.getTerritoriesWhereTransportCanUnload().entrySet())
		{	 
			territoriesWhereTransportCanUnloadCopy.put(currentTerritory.getKey(), currentTerritory.getValue());
		}	
	}
	
	public HashMap<String, Territory> getTerritoriesWhereTransportCanUnloadCopy()
	{
		return territoriesWhereTransportCanUnloadCopy;
	}
	
	public void setCanTransportMove(boolean canTransportMove)
	{
		this.canTransportMove = canTransportMove;
	}
	
	public boolean getCanTransportMove()
	{
		return canTransportMove;
	}
	
	public void setRoute(List<Territory> territoriesOfCurrentRoute)
	{
		currentRoute = new ArrayList<Territory>();
		
		for (Territory t:territoriesOfCurrentRoute)
		{
			currentRoute.add(t);
		}
	}
	
	public void clearRoute()
	{
		currentRoute = null;
	}
	
	public List<Territory> getRoute()
	{
		return currentRoute;
	}
	
	public String toString()
	{
		return unitObjectOfTransport.getID().toString();
	}
}